#include <complex>
#include <cstdlib>
#include "vmc.h"
#include <stdio.h>

void config::CreateConfiguration( int ns, int np ) 
{
	nsites = ns;
	npairs = np;

	plist = new int [2*npairs];
	latocc = new int [nsites];
	whichpair = new int [2*nsites];
	PsiInv = new double [npairs * npairs]; 

	modified = true;
}

config::config( )
{
	nsites = 0;
	npairs = 0;
	energy = 0.0;
	kinenergy = 0.0;
	othenergy = 0.0;
	modified = false;
}

config::~config( )
{
	delete plist;
	delete latocc;
	delete whichpair;
	delete PsiInv;
}

config::config( const config& from ) 
{
	int i;

	CreateConfiguration( from.nsites, from.npairs );

	for ( i = 0; i < 2 * npairs; i++ )
	{
		plist[i] = from.plist[i];
	}
	for ( i = 0; i < nsites; i++ )
	{
		latocc[i] = from.latocc[i];
	}
	for ( i  = 0; i < 2 * nsites; i++ )
	{	
		whichpair[i] = from.whichpair[i];
	}
	for ( i = 0; i < npairs * npairs; i++ )
	{
		PsiInv[i] = from.PsiInv[i];
	}

	energy = from.energy;
	kinenergy = from.kinenergy;
	othenergy = from.othenergy;
	modified = from.modified;
}

void config::CopyConfiguration( const config& from ) 
{
	int i;

	nsites = from.nsites;
	npairs = from.npairs;

	for ( i = 0; i < 2 * npairs; i++ )
	{
		plist[i] = from.plist[i];
	}
	for ( i = 0; i < nsites; i++ )
	{
		latocc[i] = from.latocc[i];
	}
	for( i = 0; i < 2 * nsites; i++ )
	{
		whichpair[i] = from.whichpair[i];
	}
	for ( i = 0; i < npairs * npairs; i++ )
	{
		PsiInv[i] = from.PsiInv[i];
	}
	
	energy = from.energy;
	kinenergy = from.kinenergy;
	othenergy = from.othenergy;
	modified = from.modified;
}

std::complex<double> config::DetPByDet( int ipair, int spin, int newsite, double *pairfn ) 
{
	double ans;
	int i, othsite;

        fprintf
          (stderr,
           "This code needs to be updated for the new latocc state names.\n");
        exit(1);

	ans = 0.0;

	if ( spin == UP )
	{
		for ( i = 0; i < npairs; i++ ) 
		{
			othsite = plist[i*2+0]; // 0 for DN
			ans = ans + pairfn[newsite*nsites+othsite] * PsiInv[i*npairs+ipair];
		}
	}
	else if ( spin == DN ) 
	{
		for ( i = 0; i < npairs; i++ ) 
		{
			othsite = plist[i*2+1]; // 1 for UP
			ans = ans + PsiInv[ipair*npairs+i] * pairfn[othsite*nsites+newsite];
		} 
	}
	else 
	{
		miscutils::InfoPrint( "****ERROR: Unknown spin in DetPByDet" );
		exit( 0 );
	}
	return ans;
}

void config::UpdateConfig( int ipair, int spin, int newsite, double *pairfn, std::complex<double> dpbdt ) 
{
	double fac;
	int i, j;
	int oldsite, othsite, ispin;

	// First, update the pair list and lattice occupancy
	ispin = ( spin + 1 ) / 2;
	oldsite = plist[ipair*2+ispin];
	plist[ipair * 2 + ispin] = newsite;
	whichpair[oldsite*2+ispin] = 0;
	if ( latocc[oldsite] == BT ) 
	{
		latocc[oldsite] = -spin;
	}
	else
	{
		latocc[oldsite] = HL;
	}
	if ( latocc[newsite] == -spin ) 
	{
		latocc[newsite] = BT;
	}
	else 
	{
		latocc[newsite] = spin;
	}
	whichpair[newsite*2+ispin] = ipair;

	// Now, update PsiInv
	if ( spin == UP )
	{
		// First, update the ipair-th column of PsiInv
		for ( i = 0; i < npairs; i++ )
		{
			PsiInv[i*npairs+ipair] = PsiInv[i*npairs+ipair] / real( dpbdt );
		}
		// Update the other columns
		for ( j = 0; j < npairs; j++ ) 
		{
			if ( j != ipair ) 
			{
				fac = 0.0;
				for ( i = 0; i < npairs; i++ )
				{
					ispin = ( 1 - spin ) / 2;
					othsite = plist[i*2+ispin];
					fac = fac + PsiInv[i*npairs+j] * pairfn[newsite*nsites+othsite];
				}
				for ( i = 0; i < npairs; i++ )
				{
					PsiInv[i*npairs+j] = PsiInv[i*npairs+j] - fac * PsiInv[i*npairs+ipair];
				}
			}
		}
	}
	else if ( spin == DN ) 
	{
		// First, update the ipair-th row of PsiInv       
		for ( i = 0; i < npairs; i++ )
		{
			PsiInv[ipair*npairs+i] = PsiInv[ipair*npairs+i] / real( dpbdt );
		}
		// Update the other rows
		for ( j = 0; j < npairs; j++ )
		{
			if ( j != ipair )
			{
				fac = 0.0;
				for ( i = 0; i < npairs; i++ ) 
				{
					ispin = ( 1 - spin ) / 2;
					othsite = plist[i*2+ispin];
					fac = fac + PsiInv[j*npairs+i] * pairfn[othsite*nsites+newsite];
				}
				for ( i = 0; i < npairs; i++ )
				{
		  			PsiInv[j*npairs+i] = PsiInv[j*npairs+i] - fac * PsiInv[ipair*npairs+i];
				}
			}
		}
	}
	else 
	{
		miscutils::InfoPrint( "****ERROR: Unknown spin in UpdateConfiguration" );
		exit( 0 );
	}
	modified = true;
}

// Overloaded function
void config::UpdateConfig( int ipair, int spin,	int newsite, double *pairfn )
{
	double fac;
        int i, j;
	std::complex<double> dpbdt;
	int oldsite, othsite, ispin;

	// Calculate ratio of determinants first
	dpbdt = DetPByDet( ipair, spin, newsite, pairfn );

	// First, update the pair list and lattice occupancy
	ispin = ( spin + 1 ) / 2;
	oldsite = plist[ipair*2+ispin];
	plist[ipair*2+ispin] = newsite;
	whichpair[oldsite*2+ispin] = 0;
	if ( latocc[oldsite] == BT )
	{
		latocc[oldsite] = -spin;
	}
	else
	{
		latocc[oldsite] = HL;
	}

	if ( latocc[newsite] == -spin ) 
	{
		latocc[newsite] = BT;
	}
	else 
	{
		latocc[newsite] = spin;
	}
	whichpair[newsite*2+ispin] = ipair;

	// Now, update PsiInv
	if ( spin == UP )
	{
		// First, update the ipair-th column of PsiInv
		for ( i = 0; i < npairs; i++ )
		{
			PsiInv[i*npairs+ipair] = PsiInv[i*npairs+ipair] / real( dpbdt );
		}
		// Update the other columns
		for ( j = 0; j < npairs; j++ )
		{
			if ( j != ipair ) 
			{
				fac = 0.0;
				for ( i = 0; i < npairs; i++ ) 
				{
					ispin = ( 1 - spin ) / 2;
					othsite = plist[i*2+ispin];
					fac = fac + PsiInv[i*npairs+j] * pairfn[newsite*nsites+othsite];
				}
				for ( i = 0; i < npairs; i++ )
				{
					PsiInv[i*npairs+j] = PsiInv[i*npairs+j] - fac * PsiInv[i*npairs+ipair];
				}
			}
		}
	}
	else if ( spin == DN ) 
	{
		// First, update the ipair-th row of PsiInv       
		for ( i = 0; i < npairs; i++ )
		{
			PsiInv[ipair*npairs+i] = PsiInv[ipair*npairs+i] / real( dpbdt );
		}
		// Update the other rows
		for ( j = 0; j < npairs; j++ )
		{
			if ( j != ipair )
			{
				fac = 0.0;
				for ( i = 0; i < npairs; i++ ) 
				{
					ispin = ( 1 - spin ) / 2;
					othsite = plist[i*2+ispin];
					fac = fac + PsiInv[j*npairs+i] * pairfn[othsite*nsites+newsite];
				}
				for ( i = 0; i < npairs; i++ )
				{
					PsiInv[j*npairs+i] = PsiInv[j*npairs+i] - fac * PsiInv[ipair*npairs+i];
				}
			}
		}
	}
	else 
	{
		miscutils::InfoPrint( "****ERROR: Unknown spin in UpdateConfiguration" );
		exit( 0 );
	}
	modified = true;
}
