// -*- c++ -*-
#ifndef CPU_DF_H
#define CPU_DF_H

#define __shared__

// Cache a row of pairfunc.
__attribute__((optimize("unroll-loops"))) void
Computation_Thread::s_pairfunc_flatten
(Pairfunction pairfunc, int newsite, int spin)
{
  for ( int row=0; row<npairs; row++ )
    s_psi_cache[row] = Pairfunc(newsite,PList(row,spin));
}

__attribute__((optimize("unroll-loops"))) void
Computation_Thread::s_tpairfunc_flatten
(Pairfunction pairfunc, int newsite, int spin)
{
  for ( int col=0; col<npairs; col++ )
    s_psi_cache[col] = TPairfunc(PList(col,spin),newsite);
}


__attribute__((optimize("unroll-loops"))) double
Computation_Thread::cpu_DetPByDet
(int ipair, int spin, int newsite,
 Pairfunction pairfunc, Configuration d_config)
{  
# ifdef COLLECT_CYCLE_COUNTS
  clock_t time_start;
  if ( !threadIdx.x ) time_start = clock64();
# endif

  double ans = 0.0;
  
  if ( spin == UP )
    {
      for ( int i=0; i<npairs; i++ )
	{
	  int othsite = PList(i,0); // 0 for DN
	  ans += Pairfunc(newsite,othsite) * PsiInv(i,ipair);
	}
    }
  else if ( spin == DN )
    {
      for ( int i=0; i<npairs; i++ )
	{
	  int othsite = PList(i,1); // 1 for UP
	  ans += TPairfunc(othsite,newsite) * PsiInv(ipair,i);
	}      
    }

# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x )
    {
      time_detpbydet[spin==DN?0:1] += clock64() - time_start;
      time_detpbydet_cnt++;
    }
# endif

  return ans;
}

__attribute__((optimize("unroll-loops"))) void
Computation_Thread::cpu_UpdateConfig
(int ipair, int ispin, int newsite, Pairfunction pairfunc, 
 double dpbdt, Configuration d_config)
{
  const int tid = threadIdx.x;
  const int mid = blockIdx.x;

# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  // First, update the pair list and lattice occupancy
 
  if(tid == 0) {
    const int oldsite = PList(ipair,ispin);
    PList(ipair,ispin) = newsite;
    Whichpair(oldsite,ispin) = 0;
  
    if ( dlatocc(oldsite) == BT )
      {
	dlatocc(oldsite) = 1-ispin;
      }
    else
      {
	dlatocc(oldsite) = HL;
      }
    if ( dlatocc(newsite) == 1-ispin )
      {
	dlatocc(newsite) = BT;
      }
    else
      {
	dlatocc(newsite) = ispin;
      }
   
    Whichpair(newsite,ispin) = ipair;
  }

  // Now, update PsiInv
  const double dinv = 1.0 / dpbdt;
  
  if ( ispin == UP )
    {
      s_pairfunc_flatten(pairfunc,newsite,0);

      // First, update the ipair-th column of PsiInv

      for ( int tid = 0; tid < npairs; tid++ ) PsiInv(tid,ipair) *= dinv;
      __syncthreads();

      // Update the other columns

      for ( int my_col=0; my_col<npairs; my_col++ ) 
        {
          if ( my_col == ipair ) continue;

          double fac = 0.0;
          for ( int row = 0; row<npairs; row++ )
            fac += PsiInv(row,my_col) * PsiEltRow(row);

          for ( int row = 0; row<npairs; row++ )          
            PsiInv(row,my_col) -= fac * PsiInv(row,ipair);
        }
    }
  else if ( ispin == DN )
    {
      const int ispin = 1;

      s_tpairfunc_flatten(pairfunc,newsite,ispin);

      // First, update the ipair-th row of PsiInv
      for ( int tid = 0; tid < npairs; tid++ ) PsiInv(ipair,tid) *= dinv;

      __syncthreads();
      
      for ( int my_row = 0; my_row < npairs; my_row++ )
        {
          if ( my_row == ipair ) continue;

          double fac = 0.0;

          for ( int col=0; col<npairs; col++ )
            fac += PsiInv(my_row,col) * PsiEltCol(col);

          for ( int col=0; col<npairs; col++ )
            PsiInv(my_row,col) -= fac * PsiInv(ipair,col);
        }
    }

  __syncthreads();
  
  if ( !threadIdx.x )
    {
      d_config.modified[mid] = true;
#     ifdef COLLECT_CYCLE_COUNTS
      time_update_config[ispin==DN?0:1] += clock64() - timer_start;
      time_update_config_cnt++;
#     endif
    }
}

__attribute__((optimize("unroll-loops"))) double
Computation_Thread::cpu_d_UpdateConfig_d
( int ipair, int spin, int newsite, int jpair, int isite)
{
# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  double rv = 0;

  if ( spin == UP )
    {
      s_pairfunc_flatten(pairfunc,newsite,0);
      __syncthreads();

      double dpbdt = 0;
      for ( int row=0; row<npairs; row++ )
        dpbdt += PsiInv(row,ipair) * PsiEltRow(row);
      
      for ( int my_col=0; my_col<npairs; my_col++ )
        {
          double fac = 0.0;

          for ( int row=0; row<npairs; row++ )
            fac += PsiInv(row,my_col) * PsiEltRow(row);

          rv +=
            my_col == ipair 
            ? ( PsiInv(jpair,ipair) * TPairfunc(newsite,isite) ) 
            : ( ( PsiInv(jpair,my_col) * dpbdt - fac * PsiInv(jpair,ipair) ) 
                * TPairfunc(PList(my_col,1),isite) );
        }
    }
  else
    {
      // Update the other rows
      const int ispin = 1;

      s_tpairfunc_flatten(pairfunc,newsite,ispin);

      __syncthreads();

      double dpbdt = 0;
      for ( int col=0; col<npairs; col++ )
        dpbdt += PsiInv(ipair,col) * PsiEltCol(col);

      for ( int row=0; row<npairs; row++ )
        {
          double fac = 0.0;

          for ( int col=0; col<npairs; col++ )
            fac += PsiInv(row,col) * PsiEltCol(col);

          const int othsite = PList(row,0);

          rv +=
            row == ipair 
            ? ( PsiInv(ipair,jpair) * TPairfunc(isite,newsite) ) 
            : ( ( PsiInv(row,jpair) * dpbdt - fac * PsiInv(ipair,jpair) ) 
                * Pairfunc(isite,othsite) );
        }
    }
  
# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x )
    {
      time_d_update_d[spin==DN?0:1] += clock64() - timer_start;
      time_d_update_d_cnt++;
    }
# endif
  return rv;
}

#undef __shared__

#endif
