#ifndef DEVICEFUNC_CUH
#define DEVICEFUNC_CUH

typedef int PI_Idx;

// Cache a row of pairfunc.
__device__ void
s_pairfunc_flatten(Pairfunction pairfunc, int newsite, int spin)
{
  const int min_iter = npairs / block_size;
  int row = threadIdx.x;
  for ( int i=0; i<min_iter; i++, row += block_size )
    s_psi_cache[0][row] = Pairfunc(newsite,PList(row,spin));
  if ( row < npairs )
    s_psi_cache[0][row] = Pairfunc(newsite,PList(row,spin));
}
__device__ void
s_tpairfunc_flatten(Pairfunction pairfunc, int newsite, int spin)
{
  const int min_iter = npairs / block_size;
  int col = threadIdx.x;
  for ( int i=0; i<min_iter; i++, col += block_size )
    s_psi_cache[0][col] = TPairfunc(PList(col,spin),newsite);
  if ( col < npairs )
    s_psi_cache[0][col] = TPairfunc(PList(col,spin),newsite);
}

__device__ void
sw_pairfunc_flatten(Pairfunction pairfunc, int newsite, int spin)
{
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;
  const int min_iter = npairs >> WP_LG;
  int row = lane;
  for ( int i=0; i<min_iter; i++, row += WP_SZ )
    s_psi_cache[wid][row] = Pairfunc(newsite,PList(row,spin));
  if ( row < npairs )
    s_psi_cache[wid][row] = Pairfunc(newsite,PList(row,spin));
}

__device__ void
swb_pairfunc_flatten(Pairfunction pairfunc, int newsite, int start_row)
{
  const int spin = 0;
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;
  const int row = start_row + lane;
  if ( lane < row_block_size )
    sb_psi_cache[wid][lane] =
      row < npairs ? Pairfunc(newsite,PList(row,spin)) : 0;
}

__device__ void
sw_tpairfunc_flatten(Pairfunction pairfunc, int newsite, int spin)
{
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;
  const int min_iter = npairs >> WP_LG;
  int col = lane;
  for ( int i=0; i<min_iter; i++, col += WP_SZ )
    s_psi_cache[wid][col] = TPairfunc(PList(col,spin),newsite);
  if ( col < npairs )
    s_psi_cache[wid][col] = TPairfunc(PList(col,spin),newsite);
}

__device__ double DetPByDet
( 
 int ipair, int spin, int newsite, int mid,
 double *pairfn, double *d_PsiInv, int *d_plist
 )
{  
  double ans = 0.0; int i, othsite;
  
  if ( spin == UP )
    {
      for ( i = 0; i < npairs; i++ )
	{
	  othsite = d_plist[mid*2*npairs+i*2+0]; // 0 for DN
	  ans += ( pairfn[newsite*nsites+othsite] * d_PsiInv[mid*npairs*npairs+i*npairs+ipair] );
	}
    }
  else if ( spin == DN )
    {
      for ( i = 0; i < npairs; i++ )
	{
	  othsite = d_plist[mid*2*npairs+i*2+1]; // 1 for UP
	  ans += ( d_PsiInv[mid*npairs*npairs+ipair*npairs+i] * pairfn[othsite*nsites+newsite] );
	}
    }
  return ans;
}

//
// Reduction Routine
//
// Uses dbl_buffer, so cannot be mixed with other code using dbl_buffer.
//
__device__ double
reduce_dbl_buffer(double val, int size, bool pre_written = false)
{
  // Return sum of VAL for threads 0 to SIZE-1.  
  // If PRE_WRITTEN is true, ignore VAL and use value already in dbl_buffer.

  const int tid = threadIdx.x;
  __shared__ double sum;
  __syncthreads(); // Don't write sum too early.
  if ( !pre_written )
    {
      if ( tid < size + 31 ) dbl_buffer[tid] = val;
      __syncthreads();
    }
  const bool one_warp = size <= WP_SZ;

  // Optimization-friendly, sync-free code for a single-warp block.
# define STEP(offset) if ( tid < offset )	\
    dbl_buffer[tid] += dbl_buffer[tid + offset];

  if ( tid < WP_SZ )
    {
      if ( !one_warp )
        for ( int i = tid + WP_SZ; i < size; i += WP_SZ )
          dbl_buffer[tid] += dbl_buffer[i];
      if ( size >= 16 ) STEP(16); 
      STEP(8); STEP(4); STEP(2); STEP(1);
      if ( !tid ) sum = dbl_buffer[0];
    }
  __syncthreads();
  return sum;

#undef STEP
}

__device__ double
reduce_dbl_buffer(int size)
{
  // Return sum of values already in dbl_buffer, from index 0 to size-1

  return reduce_dbl_buffer(0,size,true);
}

__device__ double
reduce_warp_dbl_buffer(double val)
{
  const int tid = threadIdx.x;
  const int lane = tid & WP_MA;

  dbl_buffer[tid] = val;

# define STEP(offset) if ( lane < offset ) \
    dbl_buffer[tid] += dbl_buffer[tid + offset];

  STEP(16); STEP(8); STEP(4); STEP(2); STEP(1);
# undef STEP

  return dbl_buffer[tid-lane];
}


__device__ double DetPByDet_parallel
( 
 int ipair, int spin, int newsite,
 Pairfunction pairfunc, Configuration d_config)
{  
# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t time_start;
  if ( !threadIdx.x ) time_start = clock64();
# endif

  double ans = 0.0;
  const int i = threadIdx.x;
  
  if ( spin == UP )
    {
      if(i<npairs)
	{
	  int othsite = PList(i,0); // 0 for DN
	  ans = Pairfunc(newsite,othsite) * PsiInv(i,ipair);
	}
    }
  else if ( spin == DN )
    {
      if(i<npairs)
	{
	  int othsite = PList(i,1); // 1 for UP
	  ans = TPairfunc(othsite,newsite) * PsiInv(ipair,i);
	}      
    }

  const double sum = reduce_dbl_buffer(ans,npairs);

# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x )
    {
      time_detpbydet[spin==DN?0:1] += clock64() - time_start;
      time_detpbydet_cnt++;
    }
# endif

  return sum;
}

__device__ void UpdateConfig_parallel
(
 int ipair, int ispin, int newsite, Pairfunction pairfunc, 
 double dpbdt, Configuration d_config)
{
  const int tid = threadIdx.x;
  const int mid = blockIdx.x;

# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  // First, update the pair list and lattice occupancy
 
  if(tid == 0) {
    const int oldsite = PList(ipair,ispin);
    PList(ipair,ispin) = newsite;
    Whichpair(oldsite,ispin) = 0;
  
    if ( dlatocc(oldsite) == BT )
      {
	dlatocc(oldsite) = 1-ispin;
      }
    else
      {
	dlatocc(oldsite) = HL;
      }
    if ( dlatocc(newsite) == 1-ispin )
      {
	dlatocc(newsite) = BT;
      }
    else
      {
	dlatocc(newsite) = ispin;
      }
   
    Whichpair(newsite,ispin) = ipair;
  }

  // Now, update PsiInv
  
  if ( ispin == UP )
    {
      s_pairfunc_flatten(pairfunc,newsite,0);

      // First, update the ipair-th column of PsiInv
      if ( tid < npairs ) PsiInv(tid,ipair) /= dpbdt;
      __syncthreads();

      // Update the other columns

      const int my_col = threadIdx.x / thd_per_col;
      const int my_1st_row = threadIdx.x % thd_per_col;
      if ( my_col < npairs && my_col != ipair )
        {
          double fac = 0.0;
          const int min_iter = npairs / thd_per_col;
          PI_Idx row = my_1st_row;
#pragma unroll 8
          for ( int i = 0;  i < min_iter;  i++, row += thd_per_col )
            fac += PsiInv(row,my_col) * PsiEltRow(row);
          if ( row < npairs )
            fac += PsiInv(row,my_col) * PsiEltRow(row);
          
          if ( thd_per_col > 1 )
            {
              if ( my_1st_row == 0 ) dbl_buffer[my_col] = fac;
              for ( int k = 1; k < thd_per_col; k++ )
                if ( my_1st_row == k ) dbl_buffer[my_col] += fac;
              fac = dbl_buffer[my_col];
            }

          row = my_1st_row;
#pragma unroll 8
          for ( int i = 0;  i < min_iter;  i++, row += thd_per_col )
            PsiInv(row,my_col) -= fac * PsiInv(row,ipair);
          if ( row < npairs )
            PsiInv(row,my_col) -= fac * PsiInv(row,ipair);
        }
    }
  else if ( ispin == DN )
    {
      const int ispin = 1;

      s_tpairfunc_flatten(pairfunc,newsite,ispin);

      // First, update the ipair-th row of PsiInv
      if ( tid < npairs ) PsiInv(ipair,tid) /= dpbdt;
      __syncthreads();

      // Update the other rows
      const int bfactor = 8;  // Blocking factor.
      const int my_1st_row = tid / bfactor;
      const int my_1st_col = tid % bfactor;
      const int nthds = block_size;  // nthds must be a mult of bfactor.
      const int rows_per_chunk = nthds / bfactor;

      for ( int my_row = my_1st_row; my_row < npairs; my_row += rows_per_chunk )
        if ( my_row != ipair )
          {
            double fac = 0.0;
            const int min_iter = npairs / bfactor;
            PI_Idx col = my_1st_col;

#pragma unroll 8
            for ( int i = 0;  i < min_iter;  i++, col += bfactor )
              fac += PsiInv(my_row,col) * PsiEltCol(col);
            if ( col < npairs )
              fac += PsiInv(my_row,col) * PsiEltCol(col);

            // Get sum of entire row.
            if ( bfactor == 8 && DBL_BUFFER_SIZE >= block_size + 4 )
              {
                // Optimize for popular value of bfactor.
                dbl_buffer[tid] = fac;
                fac = dbl_buffer[tid] = fac + dbl_buffer[tid+1];
                fac = dbl_buffer[tid] = fac + dbl_buffer[tid+2];
                dbl_buffer[tid] = fac + dbl_buffer[tid+4];
                fac = dbl_buffer[my_1st_row * bfactor];
              }
            else if ( bfactor > 1 )
              {
                if ( my_1st_col == 0 ) dbl_buffer[my_row] = fac;
                for ( int k = 1; k < bfactor; k++ )
                  if ( my_1st_col == k ) dbl_buffer[my_row] += fac;
                fac = dbl_buffer[my_row];
              }

            // Update row using sum.
            col = my_1st_col;
#pragma unroll 8
            for ( int i = 0;  i < min_iter;  i++, col += bfactor )
              PsiInv(my_row,col) -= fac * PsiInv(ipair,col);
            if ( col < npairs )
              PsiInv(my_row,col) -= fac * PsiInv(ipair,col);
          }
    }

  __syncthreads();
  
  if ( !threadIdx.x )
    {
      d_config.modified[mid] = true;
#     ifdef COLLECT_CYCLE_COUNTS
      time_update_config[ispin==DN?0:1] += clock64() - timer_start;
      time_update_config_cnt++;
#     endif
    }
}

__device__ double d_UpdateConfig_d
( int ipair, int spin, int newsite, 
  Pairfunction pairfunc, Configuration d_config,
  int jpair, int isite)
{
  const int tid = threadIdx.x;

# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  double rv = 0;

  if ( spin == UP )
    {
      const int my_col = threadIdx.x / thd_per_col;
      const int my_1st_row = threadIdx.x % thd_per_col;
      double fac = 0.0;
      s_pairfunc_flatten(pairfunc,newsite,0);
      __syncthreads();
      
      if ( my_col < npairs )
        {
          const int min_iter = npairs / thd_per_col;
          PI_Idx row = my_1st_row;
#pragma unroll 8
          for ( int i = 0;  i < min_iter;  i++, row += thd_per_col )
            fac += PsiInv(row,my_col) * PsiEltRow(row);
          if ( row < npairs )
            fac += PsiInv(row,my_col) * PsiEltRow(row);

          if ( my_1st_row == 0 ) dbl_buffer[my_col] = fac;
          if ( thd_per_col > 1 )
            {
              for ( int k = 1; k < thd_per_col; k++ )
                if ( my_1st_row == k ) dbl_buffer[my_col] += fac;
            }
        }
      __syncthreads();
      double col_val = 0;
      if ( tid < npairs )
        {
          const int my_col = tid;
          const int othsite = PList(my_col,1);
          const double fac = dbl_buffer[my_col];
          const double dpbdt = dbl_buffer[ipair];

          col_val =
            my_col == ipair 
            ? ( PsiInv(jpair,ipair) * TPairfunc(newsite,isite) ) 
            : ( ( PsiInv(jpair,my_col) * dpbdt - fac * PsiInv(jpair,ipair) ) 
                * TPairfunc(othsite,isite) );
        }
      rv = reduce_dbl_buffer(col_val,npairs);
    }
  else
    {
      // Update the other rows
      const int ispin = 1;

      const int bfactor = 8;  // Blocking factor.
      const int my_1st_row = tid / bfactor;
      const int my_1st_col = tid % bfactor;
      const int nthds = block_size;  // nthds must be a mult of bfactor.
      const int rows_per_chunk = nthds / bfactor;

      s_tpairfunc_flatten(pairfunc,newsite,ispin);

      __syncthreads();

      for ( int my_row = my_1st_row; my_row < npairs; my_row += rows_per_chunk )
          {
            double fac = 0.0;
            const int min_iter = npairs / bfactor;
            PI_Idx col = my_1st_col;

            // Compute sum of "our" row elements.
#pragma unroll 8
            for ( int i = 0;  i < min_iter;  i++, col += bfactor )
              fac += PsiInv(my_row,col) * PsiEltCol(col);
            if ( col < npairs )
              fac += PsiInv(my_row,col) * PsiEltCol(col);

            // Get sum of entire row.
            if ( bfactor == 8 && DBL_BUFFER_SIZE >= block_size + 4 )
              {
                // Optimize for popular value of bfactor.
                dbl_buffer[tid] = fac;
                fac = dbl_buffer[tid] = fac + dbl_buffer[tid+1];
                fac = dbl_buffer[tid] = fac + dbl_buffer[tid+2];
                if ( my_1st_col == 0 )
                  dbl_buffer[block_size+my_row] = fac + dbl_buffer[tid+4];
              }
            else if ( bfactor > 1 )
              {
                if ( my_1st_col == 0 ) dbl_buffer[my_row] = fac;
                for ( int k = 1; k < bfactor; k++ )
                  if ( my_1st_col == k ) dbl_buffer[my_row] += fac;
                fac = dbl_buffer[my_row];
                if ( my_1st_col == 0 )
                  dbl_buffer[block_size+my_row] = fac;
              }
            else
              {
                dbl_buffer[block_size+my_row] = fac;
              }
          }

      __syncthreads();
      double row_val = 0;

      if ( tid < npairs )
        {
          const int my_row = tid;
          const int othsite = PList(my_row,0);
          const double fac = dbl_buffer[block_size+my_row];
          const double dpbdt = dbl_buffer[block_size+ipair];

          row_val =
            my_row == ipair 
            ? ( PsiInv(ipair,jpair) * TPairfunc(isite,newsite) ) 
            : ( ( PsiInv(my_row,jpair) * dpbdt - fac * PsiInv(ipair,jpair) ) 
                * Pairfunc(isite,othsite) );
        }
      rv = reduce_dbl_buffer(row_val,npairs);
    }
  
# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x )
    {
      time_d_update_d[spin==DN?0:1] += clock64() - timer_start;
      time_d_update_d_cnt++;
    }
# endif
  return rv;
}

__device__ double d_UpdateConfig_d_warp_block
( int ipair, int newsite, 
  Pairfunction pairfunc, Configuration d_config,
  int jpair, int isite, int start_row)
{
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;
  const int spin = UP;

# ifdef COLLECT_CYCLE_COUNTS
  const bool per_warp = false;
  __shared__ clock_t timer_start[per_warp ? warp_count : 1];
  if ( per_warp ? !lane : !threadIdx.x ) timer_start[wid] = clock64(); 
# endif

  swb_pairfunc_flatten(pairfunc,newsite,start_row);

  double dpbdt_thread = 0.0;

  if ( row_block_size == 32 || lane < row_block_size )
    dpbdt_thread = bPsiInv(lane,ipair) * wbPsiEltRow(lane);

  const double dpbdt = reduce_warp_dbl_buffer(dpbdt_thread);
  const double psiinv_j_i = PsiInv(jpair,ipair);
  double rv =
    psiinv_j_i * Pairfunc(newsite,isite) * (1.0/(WP_SZ*num_row_blocks));

  for ( int col = lane;  col < npairs;  col += WP_SZ )
    {
      double fac = 0.0;

      for ( int lr=0; lr < row_block_size; lr++ )
        fac += bPsiInv(lr,col) * wbPsiEltRow(lr);

      rv +=
        ( PsiInv(jpair,col) * dpbdt - fac * psiinv_j_i )
        * TPairfunc(PList(col,1),isite);
    }

# ifdef COLLECT_CYCLE_COUNTS
  if ( per_warp && !lane )
    {
      atomicAdd((unsigned long long*)&time_d_update_dw[spin==DN?0:1],
                (unsigned long long)(clock64()-timer_start[wid]) );
      if ( !start_row )
        atomicAdd((unsigned*)&time_d_update_dw_cnt,1);
    }
  else if ( !per_warp && !threadIdx.x )
    {
      time_d_update_dw[spin==DN?0:1] += clock64()-timer_start[0];
      if ( !start_row )
        time_d_update_dw_cnt++;
    }
# endif
  return rv;
}

#define Fac(idx,col) \
  d_config.ext_storage \
  [ blockIdx.x * 2 * nsites * psiinv_stride + (idx) * psiinv_stride + (col) ]

__device__ void d_UpdateConfig_d_warp_block_2p_1
( int idx, int ipair, int newsite, 
  Pairfunction pairfunc, Configuration d_config,
  int jpair, int isite, int start_row)
{
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;

# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  swb_pairfunc_flatten(pairfunc,newsite,start_row);

  for ( int col = lane;  col < npairs;  col += WP_SZ )
    {
      double fac = start_row == 0 ? 0 : Fac(idx,col);

#pragma unroll
      for ( int lr=0; lr < row_block_size; lr++ )
        fac += bPsiInv(lr,col) * wbPsiEltRow(lr);

      Fac(idx,col) = fac;
    }

# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x ) time_d_update_dw[0] += clock64()-timer_start;
# endif
}

__device__ double d_UpdateConfig_d_warp_block_2p_2
( int idx, int ipair, int newsite, 
  Pairfunction pairfunc, Configuration d_config,
  int jpair, int isite, int start_row)
{
  const int tid = threadIdx.x;
  const int lane = tid & WP_MA;

# ifdef COLLECT_CYCLE_COUNTS
  __shared__ clock_t timer_start;
  if ( !threadIdx.x ) timer_start = clock64(); 
# endif

  double rv1 = 0, rv2 = 0;

  for ( int col = lane;  col < npairs;  col += WP_SZ )
    {
      const double fac = Fac(idx,col);
      const double psi = TPairfunc(PList(col,1),isite);
      rv1 += PsiInv(jpair,col) * psi;
      rv2 += fac * psi;
    }
  const double rv = 
    PsiInv(jpair,ipair) * ( Pairfunc(newsite,isite) * (1.0/(WP_SZ)) - rv2 )
    + Fac(idx,ipair) * rv1;

# ifdef COLLECT_CYCLE_COUNTS
  if ( !threadIdx.x )
    {
      time_d_update_dw[1] += clock64()-timer_start;
      time_d_update_dw_cnt++;
    }
# endif
  return rv;
}



__device__ double d_UpdateConfig_d_warp
( int ipair, int spin, int newsite, 
  Pairfunction pairfunc, Configuration d_config,
  int jpair, int isite)
{
  const int tid = threadIdx.x;
  const int wid = tid >> WP_LG;
  const int lane = tid & WP_MA;

# ifdef COLLECT_CYCLE_COUNTS
  const bool per_warp = false;
  __shared__ clock_t timer_start[per_warp ? warp_count : 1];
  if ( per_warp ? !lane : !threadIdx.x ) timer_start[wid] = clock64(); 
# endif


  double rv = 0;

  if ( spin == UP )
    {
      sw_pairfunc_flatten(pairfunc,newsite,0);

      double dpbdt_thread = 0.0;
      const int min_iter = npairs >> WP_LG;
      PI_Idx row = lane;
#pragma unroll 8
      for ( int i=0; i < min_iter; i++, row += WP_SZ )
        dpbdt_thread += PsiInv(row,ipair) * wPsiEltRow(row);
      if ( row < npairs )
        dpbdt_thread += PsiInv(row,ipair) * wPsiEltRow(row);

      const double dpbdt = reduce_warp_dbl_buffer(dpbdt_thread);
      const double psiinv_j_i = PsiInv(jpair,ipair);
      rv = psiinv_j_i * Pairfunc(newsite,isite) * (1.0/WP_SZ);

      for ( int col = lane;  col < npairs;  col += WP_SZ )
        {
          double fac = 0.0;
#pragma unroll 8
          for ( PI_Idx row=0; row<npairs; row++ )
            fac += PsiInv(row,col) * wPsiEltRow(row);

          rv +=
            ( PsiInv(jpair,col) * dpbdt - fac * psiinv_j_i )
            * TPairfunc(PList(col,1),isite);
        }
    }
  else
    {
      const int ispin = 1;
      sw_tpairfunc_flatten(pairfunc,newsite,ispin);

      double dpbdt_thread = 0;
      const int min_iter = npairs >> WP_LG;
      PI_Idx col = lane;
#pragma unroll 8
      for ( int i=0; i<min_iter; i++, col += WP_SZ )
        dpbdt_thread += PsiInv(ipair,col) * wPsiEltCol(col);
      if ( col < npairs )
        dpbdt_thread += PsiInv(ipair,col) * wPsiEltCol(col);
      const double dpbdt = reduce_warp_dbl_buffer(dpbdt_thread);

      const int bfactor = psiinv_shared ? 1 : 8;
      const int my_1st_col = tid % bfactor;
      const int my_1st_row = lane / bfactor;
      const int rows_per_warp = WP_SZ / bfactor;
      const double dpbdt_sc = dpbdt * (1.0/bfactor);
      const double psiinv_i_j = PsiInv(ipair,jpair);

      rv = psiinv_i_j*TPairfunc(isite,newsite)*(1.0/WP_SZ);

      for ( int row = my_1st_row;  row < npairs;  row += rows_per_warp )
        {
          double fac = 0.0;
          const int min_iter = npairs / bfactor;

          PI_Idx col = my_1st_col;
#pragma unroll 8
          for ( int i=0; i<min_iter; i++, col += bfactor )
            fac += PsiInv(row,col) * wPsiEltCol(col);
          if ( col < npairs )
            fac += PsiInv(row,col) * wPsiEltCol(col);
          rv +=
            ( PsiInv(row,jpair) * dpbdt_sc - fac * psiinv_i_j )
            * Pairfunc(isite,PList(row,0));

        }
    }

# ifdef COLLECT_CYCLE_COUNTS
  if ( per_warp && !lane )
    {
      atomicAdd((unsigned long long*)&time_d_update_dw[spin==DN?0:1],
                (unsigned long long)(clock64()-timer_start[wid]) );
      atomicAdd((unsigned*)&time_d_update_dw_cnt,1);
    }
  else if ( !per_warp && !threadIdx.x )
    {
      time_d_update_dw[spin==DN?0:1] += clock64()-timer_start[0];
      time_d_update_dw_cnt++;
    }
# endif
  return rv;
}

__device__ void UpdateConfig
(
	int ipair, int ispin, int newsite, int id, double *pairfn, double dpbdt, Configuration d_config
 )
{
  double fac; unsigned int i, j; int oldsite, othsite;
  
  // First, update the pair list and lattice occupancy
  oldsite = d_config.plist[id*2*npairs+ipair*2+ispin];
  d_config.plist[id*2*npairs+ipair*2+ispin] = newsite;
  d_config.whichpair[id*2*nsites+oldsite*2+ispin] = 0;
  
  if ( d_config.latocc[id*nsites+oldsite] == BT )
    {
      d_config.latocc[id*nsites+oldsite] = 1-ispin;
    }
  else
    {
      d_config.latocc[id*nsites+oldsite] = HL;
    }
  
  if ( d_config.latocc[id*nsites+newsite] == 1-ispin )
    {
      d_config.latocc[id*nsites+newsite] = BT;
    }
  else
    {
      d_config.latocc[id*nsites+newsite] = ispin;
    }
  d_config.whichpair[id*2*nsites+newsite*2+ispin] = ipair;
 
  double invdpbdt = 1/dpbdt;
  // Now, update PsiInv
  if ( ispin == UP )
    {
      
      // First, update the ipair-th column of PsiInv
      for ( i = 0; i < npairs; i++ )
	{
	  d_config.PsiInv[id*npairs*npairs+i*npairs+ipair] *= ( invdpbdt );
	}
      
      // Update the other columns
      for ( j = 0; j < npairs; j++ )
	{
	  if ( j != ipair )
	    {
	      fac = 0.0;
	      for ( i = 0; i < npairs; i++ )
		{
                  const int ospin = DN;
		  othsite = d_config.plist[id*2*npairs+i*2+ospin];
		  fac = fac + d_config.PsiInv[id*npairs*npairs+i*npairs+j] * pairfn[newsite*nsites+othsite];
		}
	      for ( i = 0; i < npairs; i++ )
		{
		  d_config.PsiInv[id*npairs*npairs+i*npairs+j] = d_config.PsiInv[id*npairs*npairs+i*npairs+j] - fac * d_config.PsiInv[id*npairs*npairs+i*npairs+ipair];
		}
	    }
	}
    }
  else if ( ispin == DN )
    {
      // First, update the ipair-th row of PsiInv
      for ( i = 0; i < npairs; i++ )
	{
	  d_config.PsiInv[id*npairs*npairs+ipair*npairs+i] *= ( invdpbdt );
	}
      // Update the other rows
      for ( j = 0; j < npairs; j++ )
	{
	  if ( j != ipair )
	    {
	      fac = 0.0;
	      for ( i = 0; i < npairs; i++ )
		{
                  const int ospin = UP;
		  othsite = d_config.plist[id*2*npairs+i*2+ospin];
		  fac = fac + d_config.PsiInv[id*npairs*npairs+j*npairs+i] * pairfn[newsite*nsites+othsite];
		}
	      for ( i = 0; i < npairs; i++ )
		{
		  d_config.PsiInv[id*npairs*npairs+j*npairs+i] = d_config.PsiInv[id*npairs*npairs+j*npairs+i] - fac * d_config.PsiInv[id*npairs*npairs+ipair*npairs+i];
		}
	    }
	}
    }
  else
    {
      //miscutils::InfoPrint("****ERROR: Unknown spin in UpdateConfiguration");
      //exit( 0 );
    }
  d_config.modified[id] = true;
}

#endif
