// Routines for dynamic compilation of CUDA code.

#include <assert.h>
#include <bzlib.h>
#include <unistd.h>
#include <fstream>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <stdarg.h>
#include "dynamic-compile.h"
#include "pstring.h"

static void
dc_fatal
(int line, const char *file, const char *fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  fprintf(stderr, "\nInternal Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
}

Dynamic_Compile::Dynamic_Compile()
{
  init_called = false;
  mi = NULL;
  devidx = -1;
  cu_context = NULL;
  compile_cucc_target = NULL;
  required_cucc_options = "";
  user_cucc_options = "";
  dynamic_compile_cucc_target = NULL;
}

void
Dynamic_Compile::init_device()
{
  if (init_called)
    return;

  init_called = true;
  myproc = 0;

  devidx = 0;

  if ( devidx == - 1 )
    {
      dc_fatal
        (__LINE__,__FILE__,
         "Attempt to initialize dynamic compilation without a\n"
         "GPU assigned.");
      return;
    }

  CE( cuInit(0) );
  CE( cuDeviceGet(&cu_device, devidx) );
  CE( cuCtxCreate(&cu_context, 0, cu_device) );

  CE( cuDeviceComputeCapability(&cc_major, &cc_minor, cu_device) );

  pStringF cucc_target_default
    ("--gpu-architecture=sm_%d%d", cc_major, cc_minor);

  if ( dynamic_compile_cucc_target
       && strlen(dynamic_compile_cucc_target) )
    compile_cucc_target = dynamic_compile_cucc_target;
  else
    compile_cucc_target = cucc_target_default.dup();

  printf("Device %d successfully initialized.\n", devidx);
}

void
Dynamic_Compile::cucc_options_set(const char* options)
{
  user_cucc_options = options;
}

bool
Dynamic_Compile::module_get
(const string module_name_p, const string src_buildtime_path_p)
{
  if ( devidx == -1 )
    dc_fatal
      ( __LINE__,__FILE__,
        "Internal Error: Dynamic GPU compilation attempted without a\n"
        "GPU assigned.");

  if ( !mi ) mi = new Module_Info();
  if ( mi->inited ) return false;
  mi->inited = true;
  mi->dm = this;
  mi->module_name = module_name_p;
  mi->src_buildtime_path = src_buildtime_path_p;
  mi->compile_reg_limit = 0;
  return true;
}

void
Dynamic_Compile::module_src_set
(int src_uncomp_chars, int src_comp_chars, char *src_words)
{
  if ( mi->src_buildtime_path == src_unpacked_buildtime_path )
    {
      mi->src_unpacked_path = src_unpacked_path;
      mi->unpacked_out_files_base = unpacked_out_files_base;
      return;
    }

  char* const userid = cuserid(NULL);
  pStringF dc_temp_dir("/tmp/vmc-%s",userid);
  mkdir(dc_temp_dir.s,0700);
  pStringF file_path_no_ext
    ("%s/%s", dc_temp_dir.s, mi->module_name.c_str());
  pStringF curr_unpacked_src_path("%s-%03d-src.cup",file_path_no_ext.s,myproc);

  char* const out_buffer = (char*) malloc(src_uncomp_chars);
  const int data_size_char = src_comp_chars;
  unsigned int uncomp_size_arg = src_uncomp_chars;

  const int rv =
    BZ2_bzBuffToBuffDecompress
    (out_buffer,&uncomp_size_arg,src_words,data_size_char,0,0);
  assert( rv == BZ_OK );
  const int fd = open(curr_unpacked_src_path.s,
                      O_WRONLY | O_CREAT | O_TRUNC, 0600);
  if ( fd == -1 )
    dc_fatal
      (__LINE__,__FILE__,
       "Could not open %s for output because %s\n",
       curr_unpacked_src_path.s, strerror(errno));

  write(fd, out_buffer, src_uncomp_chars);
  close(fd);
  free(out_buffer);

  src_unpacked_buildtime_path = mi->src_buildtime_path;
  src_unpacked_path = curr_unpacked_src_path.s;
  unpacked_out_files_base = file_path_no_ext.s;
  mi->src_unpacked_path = src_unpacked_path;
  mi->unpacked_out_files_base = unpacked_out_files_base;
}

void
Dynamic_Compile::module_vals_set_or_reset
(const string target, const string replacement, bool reset)
{
  assert( mi->inited );
  assert( reset == ( mi->lookup(target) != "" ) );

  // Don't write a zero-length string because that is used to indicate
  // a missing replacement target.
  mi->replacements[target] = replacement == "" ? " " : replacement;
}

void
Dynamic_Compile::module_vals_set(const string target)
{
  module_vals_set_or_reset(target,mi->convenience_replacement.s,false);
  mi->convenience_replacement = "";
}

void
Dynamic_Compile::module_vals_insert(const char *assignment)
{
  pStringF a("const %s;",assignment);
  mi->convenience_replacement += a;
}

void
Dynamic_Compile::module_vals_insert(const char *var, bool value)
{
  pStringF assign("bool %s = %d",var,value);
  module_vals_insert(assign.s);
}

void
Dynamic_Compile::module_vals_insert(const char *var, int value)
{
  pStringF assign("int %s = %d",var,value);
  module_vals_insert(assign.s);
}

void
Dynamic_Compile::module_vals_insert(const char *var, double value)
{
  pStringF assign("double %s = %12g",var,value);
  module_vals_insert(assign.s);
}

void
Dynamic_Compile::module_vals_insert(const char *var, double* value)
{
  pStringF assign("double* const %s = (double*)%zd;",var,value);
  mi->convenience_replacement += assign;
}

CUmodule
Dynamic_Compile::module_load()
{
  assert( mi->inited );
  const bool use_embedded = mi->src_unpacked_path.length();
  string src_path =
    use_embedded ? mi->src_unpacked_path : mi->src_buildtime_path + ".cup";
  pStringF out_files_base
    ("%s-%03d",
     use_embedded
     ? mi->unpacked_out_files_base.c_str() : mi->src_buildtime_path.c_str(),
     myproc);
  pString dynamic_cubin_file_path = out_files_base + ".cubin";
  pString cup_out_file_path = out_files_base + ".cup";
  pString sass_out_file_path = out_files_base + ".sass";

  ifstream raw_source(src_path.c_str());
  if ( !raw_source.good() )
    dc_fatal
      (__LINE__,__FILE__,
       "Could not open %s for input.\n", src_path.c_str());

  FILE* const processed_source = fopen(cup_out_file_path,"w");
  if ( !processed_source )
    dc_fatal
      (__LINE__,__FILE__,
       "Could not open %s for output.\n", cup_out_file_path.s);

  int target_found_count = 0;
  string start_marker = "QQQ_start_marker_";
  string end_marker = "QQQ_end_marker";

  // Read (preprocessed) source code for kernel, substitute actual
  // variables for certain variables, write result to new file.
  //
  string line = "";
  while ( !raw_source.eof() )
    {
      if ( line.length() == 0 )
        getline( raw_source, line );

      // Look for substitution marker on this line.
      //
      const size_t m_start_pos = line.find(start_marker);
      if ( m_start_pos == string::npos )
        {
          // If not found write the line unmodified.
          fprintf(processed_source,"%s\n", line.c_str());
          line = "";
          continue;
        }

      // Extract the key for this substitution from the source
      // file and use it to look up the replacment text.
      //
      const size_t key_start_pos = m_start_pos + start_marker.length();
      const size_t first_semi = line.find_first_of(";");
      const size_t key_end_pos = line.find_first_of(" =",key_start_pos);
      string key = line.substr(key_start_pos,key_end_pos-key_start_pos);
      string repl = mi->lookup(key);
      if ( repl == "" )
        dc_fatal(__LINE__,__FILE__,
                   "Internal Error: Could not find replacement target %s\n"
                   "in file %s (generated from CaKernel Code).\n",
                   key.c_str(), src_path.c_str());
      const size_t m_end_pos_raw = line.find(end_marker,key_end_pos);

      // If true, replace the next line with the replacment text; if
      // false, replace the text between the start and end marker
      // (which must be on this line) with the replacement text.
      //
      const bool replace_next_line =
        first_semi > key_start_pos && m_end_pos_raw == string::npos;

      if ( target_found_count == 0 )
        {
          // Anything that needs to be inserted first.
        }
      target_found_count++;
      if ( replace_next_line )
        {
          fprintf(processed_source,"%s\n",line.c_str());
          getline( raw_source, line );
          fprintf(processed_source,"%s\n", repl.c_str());
          line = "";
        }
      else
        {
          const size_t pos_end = line.find_first_of(";",m_end_pos_raw) + 1;
          const size_t start_marker_end =
            line.find_first_of(";",key_start_pos) + 1;
          fprintf(processed_source,"%s\n",
                  line.substr(0,start_marker_end).c_str());
          fprintf(processed_source,"%s\n", repl.c_str());
          line = line.substr(pos_end);
        }
    }

  raw_source.close();
  fclose(processed_source);

  assert( target_found_count >= 1 );

  // Compile edited preprocessed file.
  //
  // Note: -Xcudafe -w suppresses warnings, which are emitted undeservedly.
  // L1 cache local only  -Xptxas -dlcm=cg

  pStringF cmd
    ("nvcc --cubin %s %s %s -Xcudafe -w '%s' -o '%s' ",
     required_cucc_options,
     compile_cucc_target, user_cucc_options.c_str(),
     cup_out_file_path.s, dynamic_cubin_file_path.s);

  if ( mi->compile_reg_limit )
    cmd.sprintf("--maxrregcount %d ",mi->compile_reg_limit);

  const int rv = system(cmd.s);
  if ( rv == -1 || WEXITSTATUS(rv) != 0 )
    dc_fatal
      (__LINE__,__FILE__,
       "Problem compiling using command %s.\n",
       cmd.s);

  // Generate assembly code (for performance tuning).
  //
  pStringF cmd_asm
    ("echo '# %s ' > '%s'; cuobjdump -sass '%s' >> '%s' &",
     cmd.s, sass_out_file_path.s,
     dynamic_cubin_file_path.s, sass_out_file_path.s);
  mi->cmd_generate_asm = cmd_asm.s;

  system(mi->cmd_generate_asm.c_str());

  // Pass the compiled code to the CUDA driver.
  //
  CE( cuModuleLoad(&mi->cu_module,dynamic_cubin_file_path) );
  mi->module_loaded = true;
  return mi->cu_module;
}

void
Dynamic_Compile::module_generate_assembler()
{
  if ( mi->cmd_generate_asm == "" ) return;
  system(mi->cmd_generate_asm.c_str());
  mi->cmd_generate_asm = "";
}

void
Dynamic_Compile::module_unload()
{
  if ( ! mi->module_loaded ) return;
  CE( cuModuleUnload( mi->cu_module ) );
  mi->module_loaded = false;
  for ( map<string,Function_Info>::iterator fiter = mi->functions.begin();
        fiter != mi->functions.end(); fiter++ )
    fiter->second.inited = false;
}

Function_Info*
Dynamic_Compile::function_info_get(const string func_name)
{
  Function_Info* const fi = &mi->functions[func_name];

  if ( !fi->inited ) function_load(func_name);

  return fi;
}

CUfunction
Dynamic_Compile::function_load(const string func_name)
{
  if ( !mi->module_loaded ) module_load();
  Function_Info* const fi = &mi->functions[func_name];

  if ( !fi->inited )
    {
      fi->inited = true;
      fi->name = func_name;
      CE( cuModuleGetFunction
          (&fi->cu_function, mi->cu_module, func_name.c_str() ) );
      CE( cuFuncGetAttribute
          ( &fi->sharedSizeBytes, CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES,
            fi->cu_function ) );
      CE( cuFuncGetAttribute
          ( &fi->constSizeBytes, CU_FUNC_ATTRIBUTE_CONST_SIZE_BYTES,
            fi->cu_function ) );
      CE( cuFuncGetAttribute
          ( &fi->localSizeBytes, CU_FUNC_ATTRIBUTE_LOCAL_SIZE_BYTES,
            fi->cu_function ) );
      CE( cuFuncGetAttribute
          ( &fi->numRegs, CU_FUNC_ATTRIBUTE_NUM_REGS,
            fi->cu_function ) );
      CE( cuFuncGetAttribute
          ( &fi->maxThreadsPerBlock, CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK,
            fi->cu_function ) );
    }

  return fi->cu_function_get();
}
