// -*- c++ -*-
#ifndef DYNAMIC_COMPILE_H
#define DYNAMIC_COMPILE_H

// Routines for dynamic compilation of CUDA code.

#include <cuda.h>
#include <cuda_runtime.h>
#include <string>
#include <vector>
#include <map>
#include "drvapi_error_string.h"
#include "pstring.h"

using namespace std;

#define STR_EXPAND(t) #t
#define STR(t) STR_EXPAND(t)

#define CE(rv)                                                                \
 {                                                                            \
   const CUresult result = rv;                                                \
   if ( result != CUDA_SUCCESS )                                              \
     {                                                                        \
       fprintf(stderr,                                                        \
               "%s:%d  Error %d for CUDA Driver API call: %s\n",              \
               __FILE__, __LINE__,                                            \
               result, getCudaDrvErrorString(result));                        \
       exit(1);                                                               \
     }                                                                        \
 }

#define CU(rv)                                                                \
 {                                                                            \
   const cudaError_t result = rv;                                             \
   if ( result != cudaSuccess )                                               \
     {                                                                        \
       fprintf(stderr,                                                        \
               "%s:%d  Error %d for CUDA Runtime API call: %s\n",             \
               __FILE__, __LINE__,                                            \
               result, cudaGetErrorString(result));                           \
       exit(1);                                                               \
     }                                                                        \
 }

class Dynamic_Compile;
class Module_Info;

class Function_Info
{
public:
  Function_Info() { inited = false; }
  CUfunction cu_function_get() { return cu_function; }

  int max_threads_get()
  {
    if ( !inited ) return -1;
    int max_threads = -1;
    CE ( cuFuncGetAttribute
         ( &max_threads, CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK,
           cu_function) );
    return max_threads;
  }

  int static_shared_memory_get()
  {
    if ( !inited ) return -1;
    int max_threads = -1;
    CE ( cuFuncGetAttribute
         ( &max_threads, CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES,
           cu_function) );
    return max_threads;
  }

  string name;             // ASCII version of kernel name.
  int sharedSizeBytes, constSizeBytes, localSizeBytes;
  int numRegs, maxThreadsPerBlock;

private:
  friend class Dynamic_Compile;
  friend class Module_Info;
  CUfunction cu_function;
  bool inited;
};

class Module_Info
{
public:
  Module_Info(const Module_Info& mi){assert(false);};
  Module_Info()
  {
    inited = false;
    module_loaded = false;
    dm = NULL;
    src_unpacked_path = "";
    convenience_replacement = "";
  }
  const string lookup(const string s)
  {
    map<string,string>::iterator it = replacements.find(s);
    return it == replacements.end() ? "" : it->second;
  }

  friend class Dynamic_Compile;
  Dynamic_Compile *dm;
  bool inited;
  bool module_loaded;
  CUmodule cu_module;
  string module_name;
  int compile_reg_limit; // Maximum number of registers per thread.

  pString convenience_replacement;

  // Build time path to the source needed for current module. Used to
  // identify unpacked files that can be used by multiple modules
  // because they were all generated from the same source. (Build-time
  // paths may not be reachable at run time.)
  string src_buildtime_path;
  string src_unpacked_path;
  string unpacked_out_files_base;
  string cmd_generate_asm; // Command needed to generate assembler.

  map<string,string> replacements;
  map<string,Function_Info> functions;
};

class Dynamic_Compile {
 public:
  Dynamic_Compile();
  void init_device();

  void cucc_options_set(const char* options);

  // Make named module the current module. If values are needed for
  // a dynamic compile return true.
  bool module_get
  (const string module_name, const string src_file_path);

  void module_src_set
  (int src_uncomp_chars, int src_comp_chars, char *src_words);

  void module_reglimit_set(int limit)
    { mi->compile_reg_limit = limit; }

  // Specify a set of values to replace those following a line matching TARGET.
  // Throw error if TARGET already exists.
  void module_vals_set(const string target, const string replacement)
  { module_vals_set_or_reset(target, replacement, false); }
  // Throw error if TARGET doesn't already exist.
  void module_vals_reset(const string target, const string replacement)
  { module_vals_set_or_reset(target, replacement, true); }
  void module_vals_set_or_reset
  (const string target, const string replacement, bool reset);

  void module_vals_set(const string target); // Use collected string.
  void module_vals_insert(const char *assignment);
  void module_vals_insert(const char *var, bool value);
  void module_vals_insert(const char *var, int value);
  void module_vals_insert(const char *var, double value);
  void module_vals_insert(const char *var, double *value);


  // Compile and load current module.
  CUmodule module_load();
  // Generate assembler code for most recently compiled module.
  void module_generate_assembler();
  // Remove the module (presumably to be replaced with a better one).
  // If no module is currently loaded, do nothing.
  void module_unload();

  // If not already present, load named function from the current
  // module, return object describing function.
  Function_Info* function_info_get(const string func_name);
  CUfunction function_load(const string func_name);

  // Copy data on host at VAR_HOST_ADDR of length VAR_HOST_SIZE to a
  // symbol named VAR_NAME on device.
  // See macro CaCUDALib_CUDA_DM_Var_To_Device
  void var_to_device
  (const char *var_name, void *var_host_addr, size_t var_host_size);

  friend class Module_Info;

  CUdevice cu_device;
  CUcontext cu_context;
  int cc_major, cc_minor;

private:
  const char *compile_cucc_target;  // Target, including "-arch".
  const char *required_cucc_options; // Other options.
  char *dynamic_compile_cucc_target; // User target option.
  string user_cucc_options; // User compile options.
  Module_Info *mi; // Current module.
  bool init_called;
  int myproc, devidx;

  // Path to most recently unpacked source file.
  string src_unpacked_path;
  // Path to use for generated files if unpacked files are used.
  string unpacked_out_files_base;
  // Build time path to source (.cu) of most recently unpacked
  // file. Used to identify unpacked files that can be used by
  // multiple modules because they were all generated from the same
  // source. (Build-time paths may not be reachable at run time.)
  string src_unpacked_buildtime_path;

};

#endif
