#include <cstdio>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <sys/time.h> // used to accurately measure program runtime
#include <cstdlib>
#include <cuda.h>
#include <string.h>
#include <assert.h>
#include <curand_kernel.h>
#include <time.h>
#include <unistd.h>
#include "drvapi_error_string.h"
#include "dynamic-compile.h"
#include "config.hpp"

#ifndef VMC_COMPILER_DEP_CHECK
#include "k_common-embed-cup.cc"
#endif

using namespace std;

void transpose(double*,double*,int);

// Test for pthread return values.
#define CP(f) { const int rv=f; assert( rv == 0 ); }

double
time_wall_fp()
{
  struct timespec now;
  clock_gettime(CLOCK_REALTIME,&now);
  return now.tv_sec + ((double)now.tv_nsec) * 1e-9;
}

int64_t
time_wall_int64()
{
  struct timespec now;
  clock_gettime(CLOCK_REALTIME,&now);
  return int64_t(now.tv_sec) * 1000000000 + int64_t(now.tv_nsec);
}


float
randscale_float(int value)
{
  const int precision_lg = 23;
  const int tolerance = 10;  // Anything >= 1 should work.
  return double(value) * ((1<<precision_lg)-tolerance) / (1<<precision_lg);
}

// Checks for runtime call errors
#define cudaSafeCall( call ) {			\
	cudaError err = call;			\
	if ( cudaSuccess != err ){		\
        	fprintf( stderr, "%s(%i) : %s.\n", __FILE__, __LINE__, cudaGetErrorString( err ) );	\
        	exit( EXIT_FAILURE );		\
}}

// Checks for kernel launch errors
#define cudaCheckErr( errorMessage ) {		\
    cudaError_t err = cudaGetLastError( );	\
	if ( cudaSuccess != err ){		\
		fprintf( stderr, "%s(%i) : %s : %s.\n", __FILE__, __LINE__, errorMessage, cudaGetErrorString( err ) );	\
		exit( EXIT_FAILURE );		\
}}


#include "vmc.h" // links to all functions executed on the CPU
#define MAIN
#include "cudaMain.cuh"
#undef MAIN

// Info about GPU and each kernel.
//
struct GPU_Info {
  double chip_bw_Bps, chip_sp_flops, chip_dp_flops;
  double chip_clock_Hz;
  int chip_cc_per_mp, chip_dp_per_mp;
  int chip_int_per_mp; // A casually chosen typical value.
  int chip_regs_per_mp;
  static const int num_kernels_max = 4;
  int num_kernels;
  cudaDeviceProp cuda_prop;  // Properties of cuda device (GPU, cuda version).
  GPU_Info(){};
  GPU_Info(int device){ set_chip_data(device); }
  void set_chip_data(int device)
    {
      cudaSafeCall(cudaGetDeviceProperties(&cuda_prop,device));
      chip_cc_per_mp =
        cuda_prop.major == 1 ? 8 :
        cuda_prop.major == 2 ? ( cuda_prop.minor == 0 ? 32 : 48 ) :
        cuda_prop.major == 3 ? 192 : 0;
      chip_int_per_mp =
        cuda_prop.major == 1 ? 8 :
        cuda_prop.major == 2 ? ( cuda_prop.minor == 0 ? 16 : 16 ) :
        cuda_prop.major == 3 ? ( cuda_prop.minor == 0 ? 32 : 32 ) : 0;
      chip_dp_per_mp =
        cuda_prop.major == 1 ? 1 :
        cuda_prop.major == 2 ? ( cuda_prop.minor == 0 ? 16 : 4 ) :
        cuda_prop.major == 3 ? ( cuda_prop.minor == 0 ? 8 : 64 ) : 0;

      chip_regs_per_mp =
        cuda_prop.major == 1 ? 16 << 10 :
        cuda_prop.major == 2 ? 32 << 10 :
        cuda_prop.major == 3 ? 64 << 10 : 0;

      chip_clock_Hz = cuda_prop.clockRate * 1000.0;
      const double chip_mem_clock_Hz = cuda_prop.memoryClockRate * 1000.0;

      chip_bw_Bps =
        chip_mem_clock_Hz * ( cuda_prop.memoryBusWidth >> 3 );
      chip_sp_flops =
        1000.0 * chip_cc_per_mp * cuda_prop.clockRate
        * cuda_prop.multiProcessorCount;
      chip_dp_flops =
        1000.0 * chip_dp_per_mp * cuda_prop.clockRate
        * cuda_prop.multiProcessorCount;
    }
};

GPU_Info gpu_info;

void
cuda_init()
{
  // Initialize the driver API.
  CE( cuInit(0) );

  gpu_info.num_kernels = 0;

  // Get information about GPU and its ability to run CUDA.
  //
  int device_count;
  cudaGetDeviceCount(&device_count); // Get number of GPUs.
  if ( device_count == 0 )
    {
      fprintf(stderr,"No GPU found, exiting.\n");
      exit(1);
    }

  // Print information about the available GPUs.
  //
  for ( int dev=0; dev<device_count; dev++ )
    {
      GPU_Info gi(dev);
      printf
        ("GPU %d: %s @ %.2f GHz WITH %d MiB GLOBAL MEM\n",
         dev, gi.cuda_prop.name, gi.cuda_prop.clockRate/1e6,
         int(gi.cuda_prop.totalGlobalMem >> 20));

      printf
        ("GPU %d: CC: %d.%d  MP: %2d  CC/MP: %3d  TH/BL: %4d\n",
         dev, gi.cuda_prop.major, gi.cuda_prop.minor,
         gi.cuda_prop.multiProcessorCount,
         gi.chip_cc_per_mp,
         gi.cuda_prop.maxThreadsPerBlock);

      printf
        ("GPU %d: SHARED: %5d B  CONST: %5d B  # REGS: %5d\n",
         dev,
         int(gi.cuda_prop.sharedMemPerBlock), int(gi.cuda_prop.totalConstMem),
         gi.cuda_prop.regsPerBlock);

      printf
        ("GPU %d: L2: %d kiB   MEM to L2: %.1f GB/s\n",
         dev,
         gi.cuda_prop.l2CacheSize >> 10,
         gi.chip_bw_Bps * 1e-9);

      printf
        ("GPU %d: SP %.1f GFLOPS  OP/ELT %.2f   DP %.1f GFLOPS  OP/ELT %.2f\n",
         dev,
         gi.chip_sp_flops * 1e-9,
         4 * gi.chip_sp_flops / gi.chip_bw_Bps,
         gi.chip_dp_flops * 1e-9,
         8 * gi.chip_dp_flops / gi.chip_bw_Bps
         );

    }

  // Choose GPU 0 because we don't have time to provide a way to let
  // the user choose.
  //
  int dev = 0;
  cudaSafeCall(cudaSetDevice(dev));
  gpu_info.set_chip_data(dev);
  printf("Using GPU %d\n",dev);
}

class Main {
public:
  Main();
  void main();
  void cpu_run();
  void gpu_run();
  void cpu_free();
  void gpu_free(){};
  
  Dynamic_Compile dynamic_compile;
  sqlat latobj; 
  phyparams phyobj; 
  wavefunc wave;
  montecarlo mpolis;
  double *thop_ishell;
  int npairs, nsites;

  double start_time_compute;

  // Set in cpu- or gpu- specialized code.
  //
  int nMarkov;
  int nMP; // GPU: number of SM's.  CPU: Number of cores.
  int mc_p_MP;  // Markov chains per MP.
  double *h_PsiInv;
  int *h_plist;
  double *h_mean, *h_sd;
  int *h_naccept, *h_nsteps, *h_nneistep;
  Kernel_Timers *h_timers;


  bool psiinv_on_chip;
  double timer_clock_Hz;  // For the cycles collected by the cycle count code.
  double chip_clock_Hz;
  double chip_bw_Bps;

};

Main::Main()
{
  h_mean = h_sd = NULL;
  h_naccept = h_nsteps = h_nneistep = NULL;
  h_timers = NULL;
  thop_ishell = NULL;
}

int
main(int argc, char **argv)
{
  Main main;
  main.main();
  return 0;
}

void
Main::main()
{
  const double start_time_prog = time_wall_fp();

  // Note: parameter file re-read elsewhere.
  ConfigFile cfg( "Params.cfg" );

  const bool opt_cpu = cfg.getValueOfKey("usecpu",0);

  // Initialize common objects on the CPU
  latobj.Banner(); 
  phyobj.ReadPhyParams();
  latobj.ReadAndGenerateLattice(phyobj);
  mpolis.init(latobj); 
  wave.SetUpPairFunc(latobj);
  thop_ishell = new double[latobj.nneibs];
  for ( int i=0; i<latobj.nneibs; i++ )
    thop_ishell[i] = latobj.thop[latobj.ishell[i]];

  npairs = mpolis.current.npairs;
  nsites = mpolis.current.nsites;

  ofstream energy_data_output;
  const char* energy_data = "energy_data.txt";
  energy_data_output.open( energy_data );
  energy_data_output.precision( 16 );
  energy_data_output << fixed << "Markov Process ID\tMean Energy\tStandard Deviation\tAcceptance Ratio\tNeighbor Step Ratio" << endl;

  if ( opt_cpu )
    cpu_run();
  else
    gpu_run();

  const int mreq = mpolis.nMarkov_requested;
  printf("Number of Markov Chains %d (%d per MP) based on %s.\n",
         nMarkov, mc_p_MP,
         mreq > 0 ? "user-requested exact number." :
         mreq < 0 ? "user-requested multiple of MPs." :
         "automatic selection");

  const double elapsed_prog = time_wall_fp() - start_time_prog;
  const double elapsed = time_wall_fp() - start_time_compute;

  // Display timers
# ifndef COLLECT_CYCLE_COUNTS
  printf(" Cycle counts TURNED OFF, to turn on\n"
         "   define COLLECT_CYCLE_COUNTS.\n");
#else
  {
    printf(" Cycle Counts.\n");

    const int show_cnt = nMarkov == 1 ? 0 : min(0,nMarkov);
    const double to_ms = 1000.0 / timer_clock_Hz;
    const double to_ns = 1e9 / timer_clock_Hz;
    Kernel_Timers t_sum;
    bzero(&t_sum,sizeof(t_sum));
    for ( int i=0; i<mpolis.nMarkov; i++ )
      {
#       define TTOT(mem) t_sum.mem += h_timers[i].mem;
        TTOT(total);
        TTOT(findpairs); TTOT(move_eval); TTOT(update);
        TTOT(detpbydet_cnt);
        TTOT(detpbydet[0]); TTOT(detpbydet[1]);
        TTOT(update_config_cnt);
        TTOT(update_config[0]); TTOT(update_config[1]);
        TTOT(d_update_d_cnt); TTOT(d_update_d[0]); TTOT(d_update_d[1]);
        TTOT(d_update_dw_cnt); TTOT(d_update_dw[0]); TTOT(d_update_dw[1]);
        TTOT(eoc);
        TTOT(energy_exchange_term_cnt); TTOT(energy_exchange);
#       undef TTOT
      }
#   define TSCALE(mem) t_sum.mem /= mpolis.nMarkov;
    TSCALE(total);
    TSCALE(findpairs); TSCALE(move_eval); TSCALE(update);
    TSCALE(detpbydet_cnt);
    TSCALE(detpbydet[0]); TSCALE(detpbydet[1]);
    TSCALE(update_config_cnt);
    TSCALE(update_config[0]); TSCALE(update_config[1]);
    TSCALE(d_update_d_cnt); TSCALE(d_update_d[0]); TSCALE(d_update_d[1]);
    TSCALE(d_update_dw_cnt); TSCALE(d_update_dw[0]); TSCALE(d_update_dw[1]);
    TSCALE(eoc);
    TSCALE(energy_exchange_term_cnt); TSCALE(energy_exchange);

    printf(" %11s  %11s  %11s  %11s  %11s  %11s\n",
           "Total", "Move Prop",
           "Move Eval",
           "Update",
           "Kin E",
           "Exch E");
    printf(" %11s  %11s  %11s  %11s  %11s  %11s\n",
           "Cycles", "Cycles",
           "Cycles", "Cycles",
           "Cycles", "Cycles"
           );
    for ( int i=0; i<=show_cnt; i++ )
      {
        Kernel_Timers& timer = i < show_cnt ? h_timers[i] : t_sum;
        printf(" %11ld  %11ld  %11ld  %11ld  %11ld  %11ld\n",
               timer.total,
               timer.findpairs,
               timer.move_eval,
               timer.update,
               timer.eoc - timer.energy_exchange,
               timer.energy_exchange);
      }
      {
        printf(" %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms\n",
               to_ms * t_sum.total,
               to_ms * t_sum.findpairs,
               to_ms * t_sum.move_eval,
               to_ms * t_sum.update,
               to_ms * ( t_sum.eoc - t_sum.energy_exchange ),
               to_ms * t_sum.energy_exchange);
      }
      {
        const double cyc_inv = 100.0 / t_sum.total;
        printf(" %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%\n",
               cyc_inv * t_sum.total,
               cyc_inv * t_sum.findpairs,
               cyc_inv * t_sum.move_eval,
               cyc_inv * t_sum.update,
               cyc_inv * ( t_sum.eoc - t_sum.energy_exchange ),
               cyc_inv * t_sum.energy_exchange);
      }

    printf("\n %11s  %11s  %11s  %11s  %11s  %11s\n",
           "Total", "Move Prop",
           "DetPByD",
           "UpdateConf",
           "d Upd d",
           "Energy");
    printf(" %11s  %11s  %11s  %11s  %11s  %11s\n",
           "Cycles", "Cycles",
           "Cycles", "Cycles",
           "Cycles", "Cycles"
           );
    for ( int i=0; i<=show_cnt; i++ )
      {
        Kernel_Timers& timer = i < show_cnt ? h_timers[i] : t_sum;
        printf(" %11ld  %11ld  %11ld  %11ld  %11ld  %11ld\n",
               timer.total,
               timer.findpairs,
               timer.detpbydet[0]+ timer.detpbydet[1],
               timer.update_config[0]+ timer.update_config[1],
               timer.d_update_d[0] + timer.d_update_d[1],
               timer.eoc);
      }
      {
        printf(" %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms\n",
               to_ms * t_sum.total,
               to_ms * t_sum.findpairs,
               to_ms * ( t_sum.detpbydet[0] + t_sum.detpbydet[1] ),
               to_ms * ( t_sum.update_config[0] + t_sum.update_config[1] ),
               to_ms * ( t_sum.d_update_d[0] + t_sum.d_update_d[1] ),
               to_ms * t_sum.eoc);
      }
      {
        const double cyc_inv = 100.0 / t_sum.total;
        printf(" %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%  %10.3f%%\n",
               cyc_inv * t_sum.total,
               cyc_inv * t_sum.findpairs,
               cyc_inv * ( t_sum.detpbydet[0] + t_sum.detpbydet[1] ),
               cyc_inv * ( t_sum.update_config[0] + t_sum.update_config[1] ),
               cyc_inv * ( t_sum.d_update_d[0] + t_sum.d_update_d[1] ),
               cyc_inv * t_sum.eoc);
      }

    printf("\n %11s  %11s  %11s  %11s  %11s  %11s\n",
           "DetPByD Up", "DetPByD Dn",
           "d Upd d Up", "d Upd d Dn",
           "dUdw Up", "dUdw Dn"
           );

    printf(" %11s  %11s  %11s  %11s  %11s  %11s\n",
           "Cycles", "Cycles",
           "Cycles", "Cycles",
           "Cycles", "Cycles" );

    for ( int i=0; i<=show_cnt; i++ )
      {
        Kernel_Timers& timer = i < show_cnt ? h_timers[i] : t_sum;
        printf(" %11ld  %11ld  %11ld  %11ld  %11ld  %11ld\n",
               timer.detpbydet[0], timer.detpbydet[1],
               timer.d_update_d[0], timer.d_update_d[1],
               timer.d_update_dw[0], timer.d_update_dw[1]);
      }

    printf(" %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms  %8.0f ms\n",
           to_ms * t_sum.detpbydet[0], to_ms * t_sum.detpbydet[1],
           to_ms * t_sum.d_update_d[0], to_ms * t_sum.d_update_d[1],
           to_ms * t_sum.d_update_dw[0], to_ms * t_sum.d_update_dw[1]);


    const double find_pairs_cnt_inv =
      1.0 / ( mpolis.nsweeps * mpolis.npsweep );
    const double energy_cnt_inv = 1.0 / mpolis.navesweeps;

    printf("\n %11s  %12s  %6s  %12s\n",
           "E-Exch", "E-Exch", "Terms", "E-Exch");
    printf(" %11s  %12s  %6s  %12s\n",
           "Total", "/ Swp", "/ Swp", "/ Term");

    printf(" %8.0f ms  %9.0f ns  %6.1f  %9.0f ns\n",
           t_sum.energy_exchange * to_ms,
           t_sum.energy_exchange * energy_cnt_inv * to_ns,
           t_sum.energy_exchange_term_cnt * energy_cnt_inv,
           t_sum.energy_exchange / max(1,t_sum.energy_exchange_term_cnt)
           * to_ns );

    printf("\n %11s %7s %7s %7s %7s %7s %8s\n",
           "Total", "FPairs", "DtPByD", "UpdCnf", "DUpdD", "DUDw", "Energy");
    printf(" %11s %7s %7s %7s %7s %7s %8s\n",
           "Cycles",
           "Cyc/Itr", "Cy/Call", "Cy/Call", "Cy/Call",
           "Cy/Call",
           "Cyc/Itr" );

    for ( int i=0; i<=show_cnt; i++ )
      {
        Kernel_Timers& timer = i < show_cnt ? h_timers[i] : t_sum;
        printf(" %11ld  %6.0f  %6.0f  %6.0f  %6.0f  %6.0f  %7.0f cyc\n",
               timer.total,
               timer.findpairs * find_pairs_cnt_inv,
               ( timer.detpbydet[0] +
                 timer.detpbydet[1] )
               / max(1.0,double(timer.detpbydet_cnt)),
               (timer.update_config[0]+timer.update_config[1])
               / max(1.0,double(timer.update_config_cnt)),
               (timer.d_update_d[0]+timer.d_update_d[1])
               / max(1.0,double(timer.d_update_d_cnt)),
               (timer.d_update_dw[0]+timer.d_update_dw[1])
               / max(1.0,double(timer.d_update_dw_cnt)),
               timer.eoc * energy_cnt_inv);
      }

    printf(" %8.0f ms  %6.0f  %6.0f  %6.0f  %6.0f  %6.0f  %7.0f ns\n",
           to_ms * t_sum.total,
           to_ns * t_sum.findpairs * find_pairs_cnt_inv,
           to_ns * ( t_sum.detpbydet[0] + t_sum.detpbydet[1] )
           / max(1.0,double(t_sum.detpbydet_cnt)),
           to_ns * (t_sum.update_config[0]+t_sum.update_config[1])
           / max(1.0,double(t_sum.update_config_cnt)),
           to_ns * ( t_sum.d_update_d[0]+ t_sum.d_update_d[1])
           / max(1.0,double(t_sum.d_update_d_cnt)),
           to_ns * (t_sum.d_update_dw[0]+t_sum.d_update_dw[1])
           / max(1.0,double(t_sum.d_update_dw_cnt)),
           to_ns * t_sum.eoc * energy_cnt_inv);


#if 0
    // Compute data bandwidth-limited speed for d_Update_d.
    const int elt_size = sizeof(double);
    const int data_pf = nsites * elt_size;
    const int data_psiinv = npairs * npairs * elt_size;
    const double insn_fp = npairs * npairs + 2 * npairs;
    // Notes:
    //  Arithmetic to compute pf addr: 4 per load elt, 1 per unrolled iter.
    const double insn_int = 7 * npairs * npairs + 10 * npairs;
    printf("\nFor dUd: pf %d B, psiinv %d B,  int %.0f/%d,  fp %.0f/%d\n",
           data_pf, data_psiinv, insn_int, gpu_info.chip_int_per_mp,
           insn_fp,gpu_info.chip_dp_per_mp);
    printf("Maximum dUd: %.1f cyc BW,  %.1f + %.1f = %.1f cyc insn\n",
           ( rb_expected * ( data_pf + ( psiinv_on_chip ? 0 : data_psiinv ) ) )
           / chip_bw_Bps * chip_clock_Hz,
           insn_int/gpu_info.chip_int_per_mp,
           insn_fp/gpu_info.chip_dp_per_mp,
           insn_fp/gpu_info.chip_dp_per_mp +
           insn_int/gpu_info.chip_int_per_mp);
#endif

  }
# endif

  const bool e_to_stdout = true;

  double mean_energy=0.0, mean_sd = 0.0;
  double naccept_total = 0;
  double nstep_total = 0;

  for ( int i = 0; i < mpolis.nMarkov; i++ )
    {
      mean_energy += h_mean[i];
      mean_sd = mean_sd + h_sd[i];
      naccept_total += h_naccept[i];
      nstep_total += h_nsteps[i];

      energy_data_output << i << "\t" << h_mean[i] << "\t" << h_sd[i] << "\t" <<
	( ( double ) h_naccept[i] ) / ( ( double ) h_nsteps[i] ) << "\t" <<
	( ( double ) h_nneistep[0] ) / ( ( double ) h_naccept[i] ) << endl;

      if ( e_to_stdout && i < 4 )
        printf(" %2d  nsteps %5d  acc %5d  e %.9f\n",
               i, h_nsteps[i], h_naccept[i], h_mean[i]);
    }

  const double accept_rate = naccept_total / nstep_total;

  // Verify returned psi inverses.
  //
  double* const PsiInv_s = mpolis.current.PsiInv;
  int* const plist_s = mpolis.current.plist;
  double max_err = 0;
  for ( int i=0; i<nMarkov; i++ )
    {
      mpolis.current.plist = &h_plist[ 2 * npairs * i ];
      mpolis.current.PsiInv = &h_PsiInv[ npairs * npairs * i ];
      const double err = wave.VerifyPsiInv(mpolis.current);
      if ( err > max_err ) max_err = err;
    }
  mpolis.current.plist = plist_s;
  mpolis.current.PsiInv = PsiInv_s;

  if ( e_to_stdout )
    printf(" Mean E %.9f,  Std Dev %.9f\n",
           mean_energy/mpolis.nMarkov,
           mean_sd/mpolis.nMarkov);

  //cout.precision(15);
  energy_data_output << endl << endl << "Mean Energy:\t" << ( mean_energy/mpolis.nMarkov ) << endl;
  energy_data_output << "Mean Standard Deviation:\t" <<  mean_sd / mpolis.nMarkov << endl;

  energy_data_output 
    << "Program run time: " 
    << setprecision(6) << elapsed_prog << " seconds" << endl;
  energy_data_output << "Computation run time: " << elapsed << " seconds" << endl;
  energy_data_output.close( );


  const double run_time_s = elapsed * 1e-3;
  const double run_time_cyc = timer_clock_Hz * run_time_s;
  const double nsweeps = mpolis.nsweeps;
  const double nsteps = nsweeps * mpolis.npsweep;

  const double nelt_equilib_per_step =
    0.5 * npairs * ( 1 + npairs )
    + accept_rate * 0.5 * ( npairs + 2 * ( 1 + npairs ) );

  const double nelt_equilib_data_per_step =
    sizeof(double) *
    ( psiinv_on_chip
      ? 0.5 * ( 3 * nsites ) + accept_rate * 0.5 * ( 1 + 4 ) * nsites
      : ( 0.5 * ( npairs * ( 1 + npairs ) + 3 * nsites )
          + accept_rate * 0.5
          * ( npairs + 2 * ( 1 + npairs ) + (1 + 4 ) * nsites ) ) );

  const double nelt_energy_per_sweep =
    latobj.x * latobj.nneibs * npairs * npairs
    + 2 * nsites * ( 0.5 * ( 1 - latobj.x ) ) * npairs * npairs;

  const double nelt_energy_data_per_sweep =
    sizeof(double) *
    ( psiinv_on_chip
      ? ( latobj.x * latobj.nneibs * npairs * ( nsites )
          + 2 * nsites * ( 0.5 * ( 1 - latobj.x ) )
          * ( nsites ) )
      : ( latobj.x * latobj.nneibs * npairs * ( npairs + nsites )
          + 2 * nsites * ( 0.5 * ( 1 - latobj.x ) )
          * ( npairs * npairs + nsites ) ) );

  const double nelt =
    nsteps * nelt_equilib_per_step
    + nsweeps * nelt_energy_per_sweep;

  const double data_per_sweep =
    npsweep * nelt_equilib_data_per_step
    + nelt_energy_data_per_sweep;
  const double data_sweep_rate_limit =
    chip_bw_Bps / ( data_per_sweep * nMarkov );

  cout << endl << "Program run time: " 
       << setprecision(3) << elapsed_prog << " seconds,  "
       << setprecision(0) << fixed << run_time_cyc << " cyc.  Computation "
       << setprecision(3) << elapsed << "s." << endl;

  printf("Per Chip: %8.1f sweeps/s, %9.0f steps/s, %12.0f elt/s\n",
         nMarkov * nsweeps / run_time_s,
         nMarkov * nsteps / run_time_s,
         nMarkov * nelt / run_time_s );

  printf("Per Core: %8.1f sweeps/s, %9.0f steps/s, %12.0f elt/s %12.0f\n",
         mc_p_MP * nsweeps / run_time_s,
         mc_p_MP * nsteps / run_time_s,
         mc_p_MP * nelt / run_time_s, data_sweep_rate_limit );

  if ( max_err > 0.1 )
    printf("\n********** ");
  printf("Maximum psi_inv error is %f.\n",max_err);


  delete thop_ishell;
  if ( opt_cpu ) cpu_free(); else gpu_free();
  free(h_mean); free(h_sd); free(h_naccept); free(h_nsteps); free(h_nneistep);
}

void
Main::gpu_run()
{
  dynamic_compile.init_device();

  cuda_init();
  timer_clock_Hz = gpu_info.chip_clock_Hz;
  chip_clock_Hz = gpu_info.chip_clock_Hz;
  chip_bw_Bps = gpu_info.chip_bw_Bps;

  int nBlocks, nThreadsPerBlock, rand_seed;

  nMP = gpu_info.cuda_prop.multiProcessorCount;

  const int mreq = mpolis.nMarkov_requested;
  const int mreq_fixed = mreq < 0 ? -mreq : 1;

  const int regs_per_thd_est =
    gpu_info.cuda_prop.major == 3 && gpu_info.cuda_prop.minor == 5 ? 120 : 63;
  const int max_thds_if_fat =
    gpu_info.chip_regs_per_mp / regs_per_thd_est;

  const int site_idx_size = nsites < 256 ? 1 : 2;

  const int shared_misc = 214 + 256;
  const int psiinv_size = sizeof(double) * npairs * npairs;

  const int num_warps_est = ( max_thds_if_fat / mreq_fixed ) >> WP_LG;

  const int other_size_est_wo_db_wo_psi =
      npairs * sizeof(double)  // dbl_buffer extension
    + npairs * sizeof(double) // psi cache
    + latobj.nneibs * nsites * site_idx_size   // neiblist
    + npairs * ( 2 * site_idx_size ) // plist
    + nsites * ( 1 + 2 * site_idx_size ) // latocc, whichpair
    + 4 * nsites * site_idx_size // nearnp
    + shared_misc;  // Timer values, etc.

  const int shared_cap_after_db =
    ( 3 << 14 ) - sizeof(double) * max_thds_if_fat;

  const int shared_cap_wo_psi =
    shared_cap_after_db - other_size_est_wo_db_wo_psi;

  const int size_phi_wo_row_cache =
    // pairfunc cache
    npairs * num_warps_est * sizeof(double);

  const int size_row_cache_per_row =
    sizeof(double) * 
    ( num_warps_est  // psi warp cache
      + npairs       // psi inv cache
      );

  const int max_rows =
    shared_cap_wo_psi / size_row_cache_per_row;

  const int row_block_size = min(32,max_rows);

  const int size_phi_with_row_cache =
    row_block_size * size_row_cache_per_row;

  const int shared_size_with_psiinv =
    other_size_est_wo_db_wo_psi + size_phi_wo_row_cache + psiinv_size;

  const int shared_size_with_row_cache =
    other_size_est_wo_db_wo_psi + size_phi_with_row_cache;

  const bool psiinv_shared = psiinv_on_chip =
    mreq_fixed * shared_size_with_psiinv < shared_cap_after_db;

  const int shared_size_wo_db =
    psiinv_shared ? shared_size_with_psiinv : shared_size_with_row_cache;
  const int rb_limit_shared_est = shared_cap_after_db / shared_size_wo_db;

  const bool opt_row_cache = ( mpolis.tune_flags & 1 ) == 0;

  const bool use_row_cache = !psiinv_shared && opt_row_cache;
  const bool opt_energy_one_pass = 
    !use_row_cache || ( mpolis.tune_flags & 2 ) != 0;

  if ( rb_limit_shared_est < 1 )
    {
      fprintf
        (stderr,
         "** Lattice size too large to simulate, %d sites and %d pairs.\n"
         "   Choose a smaller value of L (currently %d) or x (currently %.4f)\n"
         "   or modify this program.\n"
         "   We apologize for any frustration this may have caused.\n",
         nsites, npairs, latobj.L, latobj.xnom);
      exit(1);
    }

  // Relatively conservative choice. Alternative is to re-compile
  // with a smaller value for thd_per_col if the number of threads
  // (based on register usage) is too large.
  const int max_usable_thds =
    dynamic_compile.cc_major == 3 && dynamic_compile.cc_minor == 5
    ? 512 : 512;

  const int thd_per_col_max = //max_usable_thds / npairs;
    max_usable_thds / ( mreq_fixed * npairs );
  const int thd_per_col =
    thd_per_col_max >= 8 ? 8 :
    thd_per_col_max >= 4 ? 4 :
    thd_per_col_max >= 2 ? 2 : 1;

  const int min_threads = thd_per_col * mpolis.current.npairs;
  const int thd_per_block_1 = ( min_threads + WP_SZ - 1 ) >> WP_LG << WP_LG;
  const int rb_limit_threads_est = max_usable_thds / thd_per_block_1;
  const int rb_limit_est = min(rb_limit_threads_est,rb_limit_shared_est);
  const int thd_per_block_2 = max_usable_thds / rb_limit_est;
  const bool use_max_number_of_warps = true;
  if ( use_max_number_of_warps )
    nThreadsPerBlock = thd_per_block_2 & ~ WP_MA;
  else
    nThreadsPerBlock = thd_per_block_1;

  // Set the number of Markov chains.
  nMarkov = mpolis.nMarkov =
    mreq > 0 ? mreq :
    mreq < 0 ? -mreq * nMP :
    rb_limit_est * nMP;

  mc_p_MP = ( nMarkov + nMP - 1 ) / nMP;

  printf("Estimated resident block per MP  %d by sm,  %d by regs\n",
         rb_limit_shared_est, rb_limit_threads_est);

  printf("Conf- MC %d,  T/C %d, block %d (%d wp), "
         "psiinv shared %d  row cache %d (%d rows) one pass %d\n",
         mpolis.nMarkov,
         thd_per_col, nThreadsPerBlock, nThreadsPerBlock >> 5,
         psiinv_shared,
         use_row_cache, row_block_size, opt_energy_one_pass);

  printf("Shared mem: wo phiinv %d, rows %d, ( %d | %d )\n",
         other_size_est_wo_db_wo_psi,
         row_block_size,
         shared_size_with_psiinv,
         shared_size_with_row_cache);

  // Allocate memory for device configuration data/arrays
  Configuration *d_config;
  cudaSafeCall( cudaMallocHost( &d_config, sizeof( Configuration ) ) );

  dynamic_compile.module_get("common","k_common.cu");
  dynamic_compile.module_src_set
    (cak__src_uncomp_chars, cak__src_comp_chars, (char*)&cak__src_words[0]);
  dynamic_compile.module_reglimit_set(regs_per_thd_est);

  CE( cuCtxSetSharedMemConfig( CU_SHARED_MEM_CONFIG_EIGHT_BYTE_BANK_SIZE ) );

  const int line_lg_dbl = 7-3;
  const int line_size_dbl = 1 << line_lg_dbl;
  const int psiinv_stride =
    ( npairs + line_size_dbl - 1 ) >> line_lg_dbl << line_lg_dbl;
  
  cudaSafeCall( cudaMalloc( &(d_config->energy), mpolis.nMarkov * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->kinenergy), mpolis.nMarkov * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->othenergy), mpolis.nMarkov * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->modified), mpolis.nMarkov * sizeof( bool ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->latocc), mpolis.nMarkov * mpolis.current.nsites * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->plist), mpolis.nMarkov * 2 * mpolis.current.npairs * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->whichpair), mpolis.nMarkov * 2 * mpolis.current.nsites * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->PsiInv), mpolis.nMarkov * mpolis.current.npairs * mpolis.current.npairs * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_config->p_PsiInv), mpolis.nMarkov * mpolis.current.npairs * psiinv_stride * sizeof( double ) ) );
  
  // Initialize wavefunctions for each Markov chain and transfer results to corresponding GPU memory locations
  for ( int i = 0; i < mpolis.nMarkov; i++ )
    {
      rand_seed = 13213 + 23 * i;
      wave.SetUpInitialWavefunction( mpolis.current, rand_seed );
      
      cudaSafeCall( cudaMemcpy( d_config->energy+(i), &mpolis.current.energy, sizeof( double ), cudaMemcpyHostToDevice ) );
      cudaSafeCall( cudaMemcpy( d_config->kinenergy+(i), &mpolis.current.kinenergy, sizeof( double ), cudaMemcpyHostToDevice ) );
      cudaSafeCall( cudaMemcpy( d_config->othenergy+(i), &mpolis.current.othenergy, sizeof( double ), cudaMemcpyHostToDevice ) );
      cudaSafeCall( cudaMemcpy( d_config->modified+(i), &mpolis.current.modified, sizeof( bool ), cudaMemcpyHostToDevice ) );
      
      cudaSafeCall( cudaMemcpy( d_config->latocc+(i*mpolis.current.nsites),
				mpolis.current.latocc, mpolis.current.nsites*sizeof(int), cudaMemcpyHostToDevice));
            
      cudaSafeCall( cudaMemcpy( d_config->plist+(i*2*mpolis.current.npairs), 
				mpolis.current.plist, 2*mpolis.current.npairs*sizeof( int ), cudaMemcpyHostToDevice ) );
      
      cudaSafeCall( cudaMemcpy( d_config->whichpair+(i*2*mpolis.current.nsites), 
				mpolis.current.whichpair, 2*mpolis.current.nsites*sizeof( int ), cudaMemcpyHostToDevice ) );
      
      cudaSafeCall( cudaMemcpy( d_config->PsiInv+(i*mpolis.current.npairs*mpolis.current.npairs), 
				mpolis.current.PsiInv, mpolis.current.npairs*mpolis.current.npairs*sizeof( double ), cudaMemcpyHostToDevice ) );
      
    }  
  // Determine block and thread dimensions
  nBlocks = mpolis.nMarkov;


#define CU_SYMBOL_TO_DEVICE(devname,src)                                      \
  {                                                                           \
    CUdeviceptr dptr;                                                         \
    size_t bytes;                                                             \
    CE( cuModuleGetGlobal(&dptr, &bytes, k_sysAve_m, devname) );              \
    CE( cuMemcpyHtoD(dptr, &src, bytes) );                                    \
  }

#define CU_STR_MEM_TO_DEVICE(struct,membername) \
  CU_SYMBOL_TO_DEVICE(#membername,struct.membername)

#define DC_STR_MEM_TO_DEVICE(struct,membername) \
  dynamic_compile.module_vals_insert(#membername,struct.membername);

#define DC_VAR_TO_DEVICE(var) \
  dynamic_compile.module_vals_insert(#var,var);
  
  void *d_timers;
  cudaSafeCall( cudaMalloc( &d_timers, nBlocks * sizeof(timers[0]) ) );


  DC_VAR_TO_DEVICE(psiinv_shared);
  DC_VAR_TO_DEVICE(row_block_size);
  DC_VAR_TO_DEVICE(use_row_cache);
  DC_VAR_TO_DEVICE(opt_energy_one_pass);
  DC_VAR_TO_DEVICE(thd_per_col);
  DC_VAR_TO_DEVICE(psiinv_stride);

  DC_STR_MEM_TO_DEVICE(mpolis.current,nsites);
  DC_STR_MEM_TO_DEVICE(mpolis.current,npairs);
  DC_STR_MEM_TO_DEVICE(latobj,nneibs);

  DC_STR_MEM_TO_DEVICE(mpolis,tune_degree);
  DC_STR_MEM_TO_DEVICE(mpolis,tune_flags);
  
  dynamic_compile.module_vals_insert("favor_neighbors",mpolis.neifrac > 0);
  dynamic_compile.module_vals_insert("block_size",nThreadsPerBlock);
  dynamic_compile.module_vals_insert
    ("asm_use_nc",
     dynamic_compile.cc_major == 3 && dynamic_compile.cc_minor == 5 );
  dynamic_compile.module_vals_set(STR(REPL_KEY_MAIN));

  const char* const site_idx_decl =
    site_idx_size == 1
    ? "typedef unsigned char Site_Idx;"
    : "typedef unsigned short Site_Idx;";
  dynamic_compile.module_vals_set(STR(REPL_KEY_SITE_IDX),site_idx_decl);

  CUmodule k_sysAve_m = dynamic_compile.module_load();

  CU_STR_MEM_TO_DEVICE(mpolis,nbins);
  CU_STR_MEM_TO_DEVICE(mpolis,ndat);

  CU_SYMBOL_TO_DEVICE("timers",d_timers);
  CU_STR_MEM_TO_DEVICE(mpolis,nsweeps);
  CU_STR_MEM_TO_DEVICE(mpolis,nearnsets);
  CU_STR_MEM_TO_DEVICE(latobj,nelecs);
  CU_STR_MEM_TO_DEVICE(phyobj,Jex);  // Note: Unused, Jij used instead.
  CU_STR_MEM_TO_DEVICE(mpolis,npsweep);
  CU_STR_MEM_TO_DEVICE(mpolis,neqlsweeps);
  CU_STR_MEM_TO_DEVICE(mpolis,navesweeps);
  CU_STR_MEM_TO_DEVICE(mpolis,nfac);
  CU_STR_MEM_TO_DEVICE(mpolis,INVTOL);
  CU_STR_MEM_TO_DEVICE(mpolis,naveseeds);
  CU_STR_MEM_TO_DEVICE(mpolis,selfrac);
  CU_STR_MEM_TO_DEVICE(mpolis,iseedseed);
  CU_STR_MEM_TO_DEVICE(wave,nattemptmax);
  CU_SYMBOL_TO_DEVICE("Jij",phyobj.Jex);

  CUfunction k_initRand_f = dynamic_compile.function_load("k_initRand");
  CUfunction k_sysAve_f = dynamic_compile.function_load("k_sysAve");
  CUfunction k_sysStat_f = dynamic_compile.function_load("k_sysStat");

  // Print information about time_step routine.
  //
  printf("\nCUDA Routine Resource Usage:\n");

  const char* fnames[] = { "k_initRand", "k_sysAve", "k_sysStat" };

  for ( unsigned int i=0; i<sizeof(fnames)/sizeof(fnames[0]); i++ )
    {
      Function_Info* const fi =
        dynamic_compile.function_info_get(fnames[i]);
      printf("For %s:\n", fi->name.c_str());
      printf("  %6d shared, %d const, %d loc, %d regs; "
             "%d max threads per block.\n",
             fi->sharedSizeBytes,
             fi->constSizeBytes,
             fi->localSizeBytes,
             fi->numRegs,
             fi->maxThreadsPerBlock);
    }

  // Estimate number of resident blocks:

  Function_Info* const k_sysAvg_fi =
    dynamic_compile.function_info_get("k_sysAve");
  const int rb_limit_sm = ( 3 << 14 ) / k_sysAvg_fi->sharedSizeBytes;
  const int rb_limit_thds =
    gpu_info.chip_regs_per_mp / ( k_sysAvg_fi->numRegs * nThreadsPerBlock );
  const int rb_limit = min(rb_limit_sm,rb_limit_thds);
  //  const int rb_expected = min(rb_limit,mc_p_MP);

  printf("For k_sysAvg, Bl/MP limits  %d (sm),  %d (regs)\n",
         rb_limit_sm, rb_limit_thds);

  // Copy arrays related to square lattice definitions
  Lattice *d_sqlat;
  cudaSafeCall( cudaMallocHost( &d_sqlat, sizeof( Lattice ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->ix), mpolis.current.nsites * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->iy), mpolis.current.nsites * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->kx), mpolis.current.nsites * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->ky), mpolis.current.nsites * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->neiblist), latobj.nneibs * mpolis.current.nsites * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->ishell), latobj.nneibs * sizeof( int ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->thop), latobj.neibshels * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &(d_sqlat->nearnp), 4 * mpolis.current.nsites * sizeof( int ) ) );

#define CUDA_MALLOC(var,amt) \
  cudaSafeCall( cudaMalloc( &(var), (amt) * sizeof( var[0] ) ) )
#define CUDA_TO_DEV(dest,src,elts) \
  cudaSafeCall( cudaMemcpy \
    ( dest, src, (elts)*sizeof(src[0]), cudaMemcpyHostToDevice ) );

  CUDA_MALLOC(d_sqlat->thop_ishell,latobj.nneibs);
  CUDA_TO_DEV(d_sqlat->thop_ishell,thop_ishell,latobj.nneibs);

  // Temporary storage for energy exchange term computation.
  //
  const int ext_elts = psiinv_stride * 2 * mpolis.current.nsites * nMarkov;
  CUDA_MALLOC(d_config->ext_storage, ext_elts * sizeof(double) );

  cudaSafeCall( cudaMemcpy( d_sqlat->ix, latobj.ix, mpolis.current.nsites*sizeof( int ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( d_sqlat->iy, latobj.iy, mpolis.current.nsites*sizeof( int ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( d_sqlat->kx, latobj.kx, mpolis.current.nsites*sizeof( double ), cudaMemcpyHostToDevice ) );	
  cudaSafeCall( cudaMemcpy( d_sqlat->ky, latobj.ky, mpolis.current.nsites*sizeof( double ), cudaMemcpyHostToDevice ) );
	
  cudaSafeCall( cudaMemcpy( d_sqlat->neiblist, latobj.neiblist, latobj.nneibs*mpolis.current.nsites*sizeof( int ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( d_sqlat->ishell, latobj.ishell, latobj.nneibs*sizeof( int ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( d_sqlat->thop, latobj.thop, latobj.neibshels*sizeof( double ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( d_sqlat->nearnp, latobj.nearnp, 4*mpolis.current.nsites*sizeof( int ), cudaMemcpyHostToDevice ) );
  

  // Allocate memory for the pair function vector common to all Markov chain
  Pairfunction * pairfunction; 
  cudaSafeCall( cudaMallocHost( &pairfunction, sizeof( Pairfunction ) ) );
  cudaSafeCall( cudaMalloc( &pairfunction->pairfn, mpolis.current.nsites * mpolis.current.nsites * sizeof( double ) ) );
  cudaSafeCall( cudaMalloc( &pairfunction->Tpairfn, mpolis.current.nsites * mpolis.current.nsites * sizeof( double ) ) );
  
  double *Tpairfn_cpu = new double[mpolis.current.nsites*mpolis.current.nsites];
  transpose(wave.pairfn,Tpairfn_cpu,mpolis.current.nsites);
  cudaSafeCall( cudaMemcpy( pairfunction->pairfn, wave.pairfn, mpolis.current.nsites*mpolis.current.nsites*sizeof( double ), cudaMemcpyHostToDevice ) );
  cudaSafeCall( cudaMemcpy( pairfunction->Tpairfn, Tpairfn_cpu, mpolis.current.nsites*mpolis.current.nsites*sizeof( double ), cudaMemcpyHostToDevice ) );
  delete[] Tpairfn_cpu;

  // Copy the neighbor list to the GPU (common to all Markov chains )
  MonteCarlo *d_mcarlo;
  cudaSafeCall( cudaMallocHost( &d_mcarlo, sizeof( MonteCarlo ) ) );

#define CUDA_MALLOC_ZERO(ptr_var, num) \
  cudaSafeCall( cudaMalloc( &(ptr_var), num * sizeof(ptr_var[0]) ) ); \
  cudaSafeCall( cudaMemset( ptr_var, 0, num * sizeof(ptr_var[0]) ) );

  CUDA_MALLOC_ZERO(d_mcarlo->naccept, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_mcarlo->nsteps, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_mcarlo->nflip, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_mcarlo->nsingle, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_mcarlo->nneistep, 1);

  double *d_mean;
  double *d_sd;
  double *d_enearr;

  // These are zeroed to help catch bugs.
  CUDA_MALLOC_ZERO(d_mean, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_sd, mpolis.nMarkov);
  CUDA_MALLOC_ZERO(d_enearr, mpolis.nMarkov * mpolis.nbins );


  Device_Constants dc_host;
  dc_host.neifrac = mpolis.neifrac;
  dc_host.nsites_randscaler = randscale_float(mpolis.current.nsites);
  dc_host.npairs_randscaler = randscale_float(mpolis.current.npairs);
  dc_host.nneibs_randscaler = randscale_float(latobj.nneibs);
  CU_SYMBOL_TO_DEVICE("dc",dc_host);

  // Setup pseudo random number generator states
  curandState *devStates;
  cudaSafeCall( cudaMalloc( ( void** ) &devStates, nBlocks * sizeof( devStates[0] ) ) );

  // Wait for memory zeroing calls to finish.
  cudaDeviceSynchronize( );

  start_time_compute = time_wall_fp();

  {
    void *args[] = { &devStates };
    CE( cuLaunchKernel
        (k_initRand_f, nBlocks, 1, 1, 1, 1, 1, 0, NULL, args, NULL) );
  }

  // Enter the energy averaging step; perform Monte Carlo sweeps over navesweeps
  // Array to store energies for all Markov chains in averaging "bins"
  {
    void *args[] =
      { d_config, d_mcarlo, d_sqlat, pairfunction, &d_enearr, &devStates };
    CE( cuLaunchKernel
        (k_sysAve_f,
         nBlocks, 1, 1,
         nThreadsPerBlock, 1, 1,
         0, // Dynamic shared memory
         NULL, // Stream
         args, //  kernel_args,
         NULL) );
  }

  {
    void *args[] = { &d_enearr, &d_mean, &d_sd };
    CE( cuLaunchKernel
        (k_sysStat_f,
         nBlocks, 1, 1,
         1, 1, 1,
         0, // Dynamic shared memory
         NULL, // Stream
         args, //  kernel_args,
         NULL) );
  }

  cudaSafeCall( cudaDeviceSynchronize() );

  // Copy energy data from all Markov chains
  h_mean = ( double* )calloc( mpolis.nMarkov,sizeof( double ) );
  h_sd = ( double* )malloc( mpolis.nMarkov * sizeof( double ) );
  h_naccept = ( int* )malloc( mpolis.nMarkov * sizeof( int ) );
  h_nsteps = ( int* )malloc( mpolis.nMarkov * sizeof( int ) );
  h_nneistep = ( int* )malloc( sizeof( int ) );

  cudaSafeCall( cudaMemcpy( h_mean, d_mean, mpolis.nMarkov * sizeof( double ), cudaMemcpyDeviceToHost ) );
  cudaSafeCall( cudaMemcpy( h_sd, d_sd, mpolis.nMarkov * sizeof( double ), cudaMemcpyDeviceToHost ) );
  cudaSafeCall( cudaMemcpy( h_naccept, d_mcarlo->naccept, mpolis.nMarkov * sizeof( int ), cudaMemcpyDeviceToHost ) );
  cudaSafeCall( cudaMemcpy( h_nsteps, d_mcarlo->nsteps, mpolis.nMarkov * sizeof( int ), cudaMemcpyDeviceToHost ) );
  cudaSafeCall( cudaMemcpy( h_nneistep, d_mcarlo->nneistep, sizeof( int ), cudaMemcpyDeviceToHost ) );
  cudaDeviceSynchronize( ); cudaSafeCall( cudaFree( d_mean ) ); cudaSafeCall( cudaFree( d_sd ) );

#define CUDA_TO_HOST(dest,src,elts) \
  cudaSafeCall( cudaMemcpy \
    ( dest, src, (elts)*sizeof(src[0]), cudaMemcpyDeviceToHost ) );

  const int psiinv_elts = npairs * npairs * nMarkov;
  h_PsiInv = new double[psiinv_elts];
  CUDA_TO_HOST(h_PsiInv,d_config->PsiInv,psiinv_elts);
  const int plist_elts = nMarkov * 2 * npairs;
  h_plist = new int[plist_elts];
  CUDA_TO_HOST(h_plist,d_config->plist,plist_elts);

  h_timers = new Kernel_Timers[nBlocks];
  cudaSafeCall( cudaMemcpy( h_timers, d_timers, nBlocks * sizeof(*h_timers),
                            cudaMemcpyDeviceToHost ) );

  cudaSafeCall( cudaFree( d_config->plist) ); cudaSafeCall( cudaFree( d_config->latocc) );
  cudaSafeCall( cudaFree( d_config->whichpair ) ); cudaSafeCall( cudaFree( d_config->PsiInv ) );
  cudaSafeCall( cudaFree( d_config->modified ) ); cudaSafeCall( cudaFree( d_config->energy ) );
  cudaSafeCall( cudaFree( d_config->kinenergy ) ); cudaSafeCall( cudaFree( d_config->othenergy ) );
  cudaSafeCall( cudaFree( d_mcarlo->naccept ) ); cudaSafeCall( cudaFree( d_mcarlo->nflip ) );
  cudaSafeCall( cudaFree( d_mcarlo->nsingle ) ); cudaSafeCall( cudaFree( d_mcarlo->nneistep ) );
  cudaSafeCall( cudaFree( d_sqlat->ix ) ); cudaSafeCall( cudaFree( d_sqlat->iy ) );
  cudaSafeCall( cudaFree( d_sqlat->kx ) ); cudaSafeCall( cudaFree( d_sqlat->ky ) );
  cudaSafeCall( cudaFree( d_sqlat->neiblist ) ); cudaSafeCall( cudaFree( d_sqlat->ishell ) );
  cudaSafeCall( cudaFree( d_sqlat->thop ) ); cudaSafeCall( cudaFree( d_sqlat->nearnp ) );
  cudaSafeCall( cudaFree( pairfunction->pairfn ) ); cudaSafeCall( cudaFree( devStates ) );
  cudaSafeCall( cudaFree( pairfunction->Tpairfn ) );
}

void
Main::cpu_run()
{
# define ALLOC_FULL(mem,num) \
  mem = (typeof mem) malloc( (num) * sizeof(mem[0]) );

  nMP = sysconf( _SC_NPROCESSORS_ONLN );

  timer_clock_Hz = 1e9;
  chip_clock_Hz = 3e9;
  chip_bw_Bps = 16e9;

  // Set the number of Markov chains.
  const int mreq = mpolis.nMarkov_requested;
  nMarkov = mpolis.nMarkov =
    mreq == 0 ? nMP :
    mreq > 0 ? mreq : - nMP * mreq;

  mc_p_MP = ( nMarkov + nMP - 1 ) / nMP;

  printf("Running on CPU.  Total Cores:  %d.   Chains: %d\n", nMP,nMarkov);

  psiinv_on_chip = false;

  Lattice& d_sqlat = c_sqlat;

# define ALLOCPY(mem,num) ALLOC_FULL(d_sqlat.mem,num); \
  memcpy(d_sqlat.mem, latobj.mem, (num)*sizeof(latobj.mem[0]) );

  ALLOCPY(ix,nsites);
  ALLOCPY(iy,nsites);
  ALLOCPY(kx,nsites);
  ALLOCPY(ky,nsites);
  ALLOCPY(neiblist,latobj.nneibs*nsites);
  ALLOCPY(ishell,latobj.nneibs);
  ALLOCPY(thop,latobj.neibshels);
  ALLOCPY(nearnp,4*nsites);

# undef ALLOCPY

  // Allocate memory for device configuration data/arrays
  Configuration& d_config = c_config;

  const int line_lg = 6; // UPDATE
  const int line_sz = 1 << line_lg;
# define FIND_STRIDE(num,t) \
  (((num)*sizeof(t)+line_sz-1 >> line_lg << line_lg) / sizeof(t) )

  const int psiinv_stride = FIND_STRIDE(npairs,double);

# define ALLOC(mem,num) ALLOC_FULL(d_config.mem,num)

  ALLOC(energy,nMarkov);
  ALLOC(kinenergy,nMarkov);
  ALLOC(othenergy,nMarkov);
  ALLOC(modified,nMarkov);
  ALLOC(latocc,nMarkov*nsites);
  ALLOC(plist,2*nMarkov*npairs);
  ALLOC(whichpair,2*nMarkov*nsites);
  ALLOC(PsiInv,nMarkov*npairs*npairs);
  ALLOC(p_PsiInv,nMarkov*npairs*psiinv_stride);
# undef ALLOC


#define COPYI(mem,amt)                                                        \
      memcpy(&d_config.mem[i*(amt)], mpolis.current.mem,                      \
             (amt)*sizeof(d_config.mem[0]) );

# define COPYIS(mem) \
      memcpy(&d_config.mem[i], &mpolis.current.mem, sizeof(d_config.mem[0]) );

  // Initialize wavefunctions for each Markov chain and transfer results to corresponding GPU memory locations
  for ( int i = 0; i < mpolis.nMarkov; i++ )
    {
      const int rand_seed = 13213 + 23 * i;
      wave.SetUpInitialWavefunction( mpolis.current, rand_seed );

      COPYIS(energy);
      COPYIS(kinenergy);
      COPYIS(othenergy);
      COPYIS(modified);
      COPYI(latocc,nsites);
      COPYI(plist,2*npairs);
      COPYI(whichpair,2*nsites);
      COPYI(PsiInv,npairs*npairs);

#undef COPYI
    }

  d_sqlat.thop_ishell = thop_ishell;

  h_timers = timers = new Kernel_Timers[nMarkov];

  cpu_set_vars(mpolis,latobj,phyobj,psiinv_stride);

# define S(m) m = mpolis.m;
  S(nbins);
  S(ndat);
  S(nsweeps);
  S(nearnsets);
  S(npsweep);
  S(neqlsweeps);
  S(navesweeps);
  S(navesweeps);
# undef S
  nelecs = latobj.nelecs;
  Jij = Jex = phyobj.Jex;

  // Then, allocate memory for the pair function vector common to all
  // Markov chain

  Pairfunction& pairfunction = c_pairfunc;
# define ALLOC(mem,num) ALLOC_FULL(pairfunction.mem,num)
  ALLOC(pairfn,nsites*nsites);
  ALLOC(Tpairfn,nsites*nsites);
  
  memcpy(pairfunction.pairfn, wave.pairfn, 
         nsites * nsites * sizeof(wave.pairfn[0]) );
  transpose(wave.pairfn,pairfunction.Tpairfn,mpolis.current.nsites);

  MonteCarlo& d_mcarlo = c_mcarlo;

#define MALLOC_ZERO(ptr_var, num) \
  ptr_var = (typeof ptr_var) calloc(num, sizeof(ptr_var[0]) );

  MALLOC_ZERO(d_mcarlo.naccept, mpolis.nMarkov);
  MALLOC_ZERO(d_mcarlo.nsteps, mpolis.nMarkov);
  MALLOC_ZERO(d_mcarlo.nflip, mpolis.nMarkov);
  MALLOC_ZERO(d_mcarlo.nsingle, mpolis.nMarkov);
  MALLOC_ZERO(d_mcarlo.nneistep, 1);

  double *d_mean;
  double *d_sd;
  double*& d_enearr = c_enearr;

  // These are zeroed to help catch bugs.
  MALLOC_ZERO(d_mean, mpolis.nMarkov);
  MALLOC_ZERO(d_sd, mpolis.nMarkov);
  MALLOC_ZERO(d_enearr, mpolis.nMarkov * mpolis.nbins );

  dc.neifrac = mpolis.neifrac;
  dc.nsites_randscaler = randscale_float(mpolis.current.nsites);
  dc.npairs_randscaler = randscale_float(mpolis.current.npairs);
  dc.nneibs_randscaler = randscale_float(latobj.nneibs);

  start_time_compute = time_wall_fp();

  {
    pthread_t tids[nMarkov];

    for ( int i=0; i<nMarkov; i++ )
      CP(pthread_create(&tids[i], NULL, &cpu_sysAve, (void*)ptrdiff_t(i)));

    for ( int i=0; i<nMarkov; i++ )
      CP(pthread_join(tids[i],NULL));
  }

  for ( int id=0; id<nMarkov; id++ )
    {
      double sum = 0, sos = 0;

      for ( int i = 0; i < nbins; i++ )
	{
          sum += d_enearr[nbins*id+i]; 
          sos += d_enearr[nbins*id+i] * d_enearr[nbins*id+i];
	}

      d_mean[id] = sum / double( nbins ); 
      const double sosm = sos / double( nbins );
      d_sd[id] = sqrt( ( sosm - d_mean[id] * d_mean[id] ) / ( nbins - 1 ) );
    }

  h_mean = d_mean;
  h_sd = d_sd;
  h_naccept = d_mcarlo.naccept;
  h_nsteps = d_mcarlo.nsteps;
  h_nneistep = d_mcarlo.nneistep;

  h_PsiInv = d_config.PsiInv;
  h_plist =  d_config.plist;

  free(d_config.latocc);
  free(d_config.whichpair ); 
  free(d_config.modified ); free(d_config.energy );
  free(d_config.kinenergy ); free(d_config.othenergy );
  free(d_mcarlo.nflip );
  free(d_mcarlo.nsingle );
  free(d_sqlat.ix ); free(d_sqlat.iy );
  free(d_sqlat.kx ); free(d_sqlat.ky );
  free(d_sqlat.neiblist ); free(d_sqlat.ishell );
  free(d_sqlat.thop ); free(d_sqlat.nearnp );
  free(pairfunction.pairfn );
  free(pairfunction.Tpairfn );
}

void
Main::cpu_free()
{
  free(h_PsiInv);
  free(h_plist);
}

  void transpose(double* pairfn, double* transpose, int n)
  {
    const int blocksize = n;
    for (int i = 0; i < n; i += blocksize) {
      for (int j = 0; j < n; j += blocksize) {
	for (int k = i; k < i + blocksize; ++k) {
	  for (int l = j; l < j + blocksize; ++l) {
	    transpose[k + l*n] = pairfn[l + k*n];
	  }
        }
      }
    }
  }
