#include <iostream>
#include <cstdlib>
#include "vmc.h"
#include "config.hpp"
#include <fstream>
#include <iomanip>

// Generate banner
void sqlat::Banner( void )
{
	using namespace std;
	cout << "=========================================================" << endl;
	cout << "      Variational Monte Carlo of Hubbard Model           " << endl;
	cout << "                 (Tilted Square Lattice)                 " << endl;
	cout << "=========================================================" << endl;
	cout << endl;
}

// Allocate space for lattice-descriptive arrays
void sqlat::AllocateSpace( void ) 
{
	ix = new int [nsites];
	iy = new int [nsites];
	kx = new double [nsites];
	ky = new double [nsites];
	neiblist = new int [nneibs*nsites];
	ishell = new int [nneibs];
	thop = new double [neibshels];
	nearnp = new int [4*nsites];
}

// Read and generate the generic lattice
void sqlat::ReadAndGenerateLattice( phyparams& pp ) 
{
	using namespace std;

	int i, j, itemp, count, isx, iex;

	ConfigFile cfg( "Params.cfg" );		
	L = cfg.getValueOfKey<int>( "L" );
	x = cfg.getValueOfKey<double>( "x" );

	if ( L % 2 != 1 ) 
	{
		miscutils::InfoPrint( "***ERROR: L should be odd, thank you very much!" );
		exit( 0 );
	}

	// Calculate the total number of sites  
	nsites = L * L + 1 ;
	Lby2 = L / 2;

	// Calculate the number of electrons based on provided doping concentration
	nelecs = ( int ) ( ( ( double ) nsites ) * ( 1.E0 - x ) );

	ofstream outfile;
	outfile.open( "scrvb.info" );
	outfile << setw( 9 ) << setprecision ( 6 ) << x << setw( 5 ) << setprecision( 0 ) << nelecs;
	outfile.close( );

	// Ensure there are an even number electrons
	if ( nelecs % 2 != 0 ) 
	{
		nelecs = nelecs + 1;
	}
	npairs = nelecs / 2;

	// Save the input doping as nominal doping
	xnom = x;

	// Calculate actual doping concentration (number of holes per site)
	x = ( ( double ) ( nsites - nelecs ) ) / ( ( double ) nsites ); 

	// Set up neighbour info
	nneibs = 12, neibshels = 3;

	miscutils::InfoPrint( "Lattice and Doping Info" );
	miscutils::InfoPrint( "Lattice size (L)", L );
	miscutils::InfoPrint( "Number of sites", nsites );
	miscutils::InfoPrint( "Nominal hole doping", xnom );
	miscutils::InfoPrint( "Number of electrons", nelecs );
	miscutils::InfoPrint( "Actual doping (x)", x );       

	// Allocate space for lattice arrays
	AllocateSpace( );

	// Set up hopping matrix elements
	thop[0] = -pp.t;
	thop[1] = pp.tp;
	thop[2] = -pp.tpp;

	Jij = pp.Jex;

	// Generate lattice sites
	count = 0;
	for ( j = -Lby2; j <= Lby2; j++ ) 
	{
		if ( j < 0 ) 
		{
			isx = -Lby2;
			iex = Lby2;
		}
		else if ( j == 0 ) 
		{
			isx = -Lby2;
			iex = Lby2 + 1;
		}
		else if ( j > 0 )
		{
			isx = -Lby2 + 1;
			iex = Lby2 + 1;
		}

		for ( i = isx; i <= iex; i++ ) 
		{
			ix[count] = i;
			iy[count] = j;
			count++;
		}
	}

	// Set up periodic vectors for future reference
	lvec[0][0] = L;
	lvec[1][0] = -1;
	lvec[0][1] = 1;
	lvec[1][1] = L;

	FindNeighbors( );

	// Now, find near neighbour sets
	nearnsets = 2 * nsites;
	itemp = 0;
	for ( i = 0; i < nsites; i++ ) 
	{
		for ( j = 0; j < nneibs; j++ ) 
		{
			if ( ishell[j] == 0 ) 
			{
				if ( neiblist[j+nneibs*i] > i ) 
				{
					nearnp[itemp] = i;
					nearnp[itemp+1] = neiblist[j+nneibs*i];
					itemp += 2;
				}
			}
		}
	}

	if ( itemp != 2 * nearnsets )
	{
		miscutils::InfoPrint( "Problem in near neighbor sets", nearnsets );
		exit( 0 );
	}

	// Now, determine the reciprocal lattice 
	itemp = 0;
	for ( i = -( L - 1 ) / 2; i <= ( L - 1 ) / 2; i++ ) 
	{
		for ( j= -( L - 1 ) / 2; j <= ( L - 1 ) / 2; j++ ) 
		{
			kx[itemp] = 2.0 * M_PI * ( ( double ) ( i * L + j ) ) / ( ( double ) nsites );
			ky[itemp] = 2.0 * M_PI * ( ( double ) ( -i + L * j ) ) / ( ( double ) nsites );
			itemp++;
		}
	}
	kx[itemp] = M_PI;
	ky[itemp] = M_PI;
	if ( itemp != nsites - 1 ) 
	{
		miscutils::InfoPrint( "Inconsistency in the reciprocal space" );
		exit( 0 );
	}
}

void sqlat::FindNeighbors( void ) 
{
	using namespace std;

	int *ine, *inx, *in2; 
	int i, j, v[2];

	ine = new int [nsites];
	inx = new int [nsites];
	in2 = new int [nsites];

	fill_n( ine, nsites, 0 );
	fill_n( inx, nsites, 0 );
	fill_n( in2, nsites, 0 );
	fill_n( neiblist, nneibs * nsites, -1 );

	for ( i = 0; i < nsites; i++ ) 
	{
		for ( j = i + 1; j < nsites; j++ ) 
		{
                        VecDistance( i, j, v );
			if ( ( v[0] == 1 ) && ( v[1] == 0 ) ) 
			{
				ine[i] = ine[i] + 1;
				ine[j] = ine[j] + 1;
				neiblist[0+nneibs*i] = j;
				neiblist[2+nneibs*j] = i;
			}
			else if ( ( v[0] == 0 ) && ( v[1] == 1 ) ) 
			{
				ine[i] = ine[i] + 1;
				ine[j] = ine[j] + 1;
				neiblist[1+nneibs*i] = j;
				neiblist[3+nneibs*j] = i;
			}
			else if ( ( v[0] == -1 ) && ( v[1] == 0 ) ) 
			{
				ine[i] = ine[i] + 1;
				ine[j] = ine[j] + 1;
				neiblist[2+nneibs*i] = j;
				neiblist[0+nneibs*j] = i;
			}
			else if ( ( v[0] == 0 ) && ( v[1] == -1 ) ) 
			{
				ine[i] = ine[i] + 1;
				ine[j] = ine[j] + 1;
				neiblist[3+nneibs*i] = j;
				neiblist[1+nneibs*j] = i;
			}

			// Now, look for second neighbours
			else if ( ( v[0] == 1 ) && ( v[1] == 1 ) ) 
			{
				inx[i] = inx[i] + 1;
				inx[j] = inx[j] + 1;
				neiblist[4+nneibs*i] = j;
				neiblist[6+nneibs*j] = i;
			}
			else if ( ( v[0] == -1 ) && ( v[1] == 1 ) ) 
			{
				inx[i] = inx[i] + 1;
				inx[j] = inx[j] + 1;
				neiblist[5+nneibs*i] = j;
				neiblist[7+nneibs*j] = i;
			}
			else if ( ( v[0] == -1 ) && ( v[1] == -1 ) ) 
			{
				inx[i] = inx[i] + 1;
				inx[j] = inx[j] + 1;
				neiblist[6+nneibs*i] = j;
				neiblist[4+nneibs*j] = i;
			}
			else if ( ( v[0] == 1 ) && ( v[1] == -1 ) ) 
			{
				inx[i] = inx[i] + 1;
				inx[j] = inx[j] + 1;
				neiblist[7+nneibs*i] = j;
				neiblist[5+nneibs*j] = i;
			}
			// Now, look for third neighbours
			else if ( ( v[0] == 2 ) && ( v[1] == 0 ) ) 
			{
				in2[i] = in2[i] + 1;
				in2[j] = in2[j] + 1;
				neiblist[8+nneibs*i] = j;
				neiblist[10+nneibs*j] = i;
			}
			else if ( ( v[0] == 0 ) && ( v[1] == 2 ) ) 
			{
				in2[i] = in2[i] + 1;
				in2[j] = in2[j] + 1;
				neiblist[9+nneibs*i] = j;
				neiblist[11+nneibs*j] = i;
			}
			else if ( ( v[0] == -2 ) && ( v[1] == 0 ) ) 
			{
				in2[i] = in2[i] + 1;
				in2[j] = in2[j] + 1;
				neiblist[10+nneibs*i] = j;
				neiblist[8+nneibs*j] = i;
			}
			else if ( ( v[0] == 0 ) && ( v[1] == -2 ) ) 
			{
				in2[i] = in2[i] + 1;
				in2[j] = in2[j] + 1;
				neiblist[11+nneibs*i] = j;
				neiblist[9+nneibs*j] = i;
			}
		}
	}


	// Check if everything is okay
	for ( i = 0; i < nsites; i++ ) 
	{
		if ( ( ine[i] != 4 ) || ( inx[i] != 4 ) || ( in2[i] != 4 ) ) 
		{
			miscutils::InfoPrint( "***ERROR: Did not find enough neighbours" );
			exit( 0 );
		}
		for ( j = 0; j < nneibs; j++ ) 
		{
			if ( ( neiblist[j+nneibs*i] == -1 ) )
			{
				cout << i << " " << j << endl;
				miscutils::InfoPrint( "***ERROR: Neighbour list not ok" );
				exit( 0 );
			}
		}	
	}

	// Define neighbour shells
	for ( j = 0; j < nneibs; j++ ) 
	{
		if ( j < 4 ) 
		{
			ishell[j] = 0;
		}
		else if ( j < 8 ) 
		{
			ishell[j] = 1;
		} 
		else ishell[j] = 2;
	}
}

void PrintNeighbors( );

int sqlat::VecDistance( int i, int j, int *v )
{
	using namespace std;

	int d2, k, ic, jc, d2t;
	int vt[2], ri[2], rj[2], rs[2], shift[2];

	if ( i == j )
	{
		d2 = 0;
		fill_n( v, 2, 0 );
		return d2;
	}

	ri[0] = ix[i];
	ri[1] = iy[i];
	rj[0] = ix[j];
	rj[1] = iy[j];

	vt[0] = rj[0] - ri[0];
	vt[1] = rj[1] - ri[1];

	d2 = vt[0] * vt[0] + vt[1] * vt[1];
	for ( k = 0; k < 2; k++ ) 
	{
		v[k] = vt[k];
	}

	for ( ic = -1; ic <= 1; ic++ ) 
	{
		for ( jc = -1; jc <= 1; jc++ ) 
		{
			if ( ( ic != 0 ) || ( jc != 0 ) ) 
			{
				shift[0] = ic * lvec[0][0] + jc * lvec[0][1];
				shift[1] = ic * lvec[1][0] + jc * lvec[1][1];
				for ( k = 0; k < 2; k++ ) 
				{
					rs[k] = rj[k] + shift[k];
					vt[k] = rs[k] - ri[k];
				}
				d2t = vt[0] * vt[0] + vt[1] * vt[1];
				if ( d2t < d2 ) 
				{
					d2 = d2t;
					for ( k = 0; k < 2; k++ ) 
					{
						v[k]  = vt[k];
					}
				}
			}
		}
	}
	return d2;
}
