#ifndef VMC_H
#define VMC_H

#include <stdint.h>
#include <string>
#include <complex>

#define DN 0
#define UP 1
#define HL 3
#define BT 2

double time_wall_fp();  // Returns seconds.
int64_t time_wall_int64();  // Returns nanoseconds.


class ConfigFile;

class phyparams 
{
public:
	// Non-dimensional (everything is divided out by t; finally, t will be 1)
	double t, tp, tpp, Jex;

	// Possibly dimensional (i.e., what is actully input)
	double tvalue, tpvalue, tppvalue, Jexvalue;

	void ReadPhyParams( void );
};

class sqlat
{
public:
	double x, xnom;
	int L, nsites, Lby2, nelecs, npairs;
	// x, y coordinate of each site
	int *ix, *iy;
	int lvec[2][2];

	// kx, ky		: reciprocal lattice vector coordinates
	// nredk		: n-reduced k vectors
	double *kx, *ky;
	int nredk;

	// nneibs		: number of neighbours for each site
	// neibshels		: number of neighbour shells
	int nneibs, neibshels;

	// neiblist( i, j )	: i-th neighbour of the j-th site
	int* neiblist;

	/* ishell(i)		: the shell of the i-th neighbour 
				  (this will decide hopping matrix element) */
	int* ishell;

	/* thop(i)		: hopping matrix elements between i-th  neighbours
				  (i = 1 is nearest neighbour, i = 2 next nearest etc.) */
	double* thop;

	// Jij			: antiferromagnetic exchange
	double Jij;

	/* Stuff added for easy parallelization
	nearnsets		: number of near neighbour sets
	nearnp(i,1)		: the first site of the set, (:,2) second site of the set */
	int nearnsets;
	int* nearnp;

	// Function declarations 
	void Banner( void );
	void AllocateSpace( void );
	void ReadAndGenerateLattice( phyparams& );
	void FindNeighbors( void );
	void PrintNeighbors( );
	int VecDistance( int, int, int* );
};

class config 
{
public:
	double energy;
	double kinenergy;
	double othenergy;
	int nsites, npairs;

	/* plist(i,is)		: gives the lattice site of the ith electron with spin is
	   			  0: DN, 1: UP */
	int *plist;

	// latocc(:)		: latocc(i) = -1, 1, 0, 2 for down, up, holon, doublon at site i
	int *latocc;
	/* whichpair(ilat,spin)	: Tells which pair has an electron of spin "spin" at 
				  lattice site ilat */
	int *whichpair;
	double *PsiInv;
	bool modified;

	config( );
	config( const config& );
	void CreateConfiguration( int, int );
	void CopyConfiguration( const config& );
	std::complex<double> DetPByDet( int, int, int, double* );
	void UpdateConfig( int, int, int, double* );
	void UpdateConfig( int, int, int, double*, std::complex<double> );
	~config( );
};

class wavefunc 
{
	double ConditionNumberMin, BigNumber;
	double Delta, mu;
	int nvarp;

public:
	int nattemptmax;
	double *pairfn;
	void SetUpPairFunc( sqlat& );
	void SetUpInitialWavefunction( config&, int );
	wavefunc( );
        double VerifyPsiInv(config& conf);
        double VerifyInv
          (config& conf, double *a, double *b, const char *msg = NULL);
};

class montecarlo
{
	sqlat tilt;
	wavefunc rvb;
	//config current;
public:
        int nfac, nbins, naccept, nsingle, nflip,
                ndat, nneistep, naveseeds, iseedseed, nsteps;
        double INVTOL, neifrac, selfrac;
        bool WriteHistory;

        int nMarkov_requested; // Value of user argument. Used to set nMarkov.
  int nMarkov, nsweeps, neqlsweeps, navesweeps, npsweep, nearnsets;
	montecarlo(){};
	montecarlo( sqlat& s ) { init( s ); }
	void init( sqlat& );
	config current; // needs to be public to transfer CPU data to GPU

        int tune_degree;
        int tune_flags;
};

namespace miscutils 
{
	void InfoPrint( void );
	void InfoPrint( std::string );
	void InfoPrint( std::string, double );
	void InfoPrint( std::string, char );
	void InfoPrint( std::string, int );
	void InitializeRandomNumGenerator( int );
	int RandomInteger( int );
	double RandomReal( void );
	void CalculateAverages( int, double*, double&, double& );
};

#endif
