#include <cmath>
#include <complex>
#include <cstdlib>
#include <iomanip>
#include "vmc.h"
#include "config.hpp"
#include <fstream>
using namespace std;

#define tol 1.e-15

extern "C" double zlange_( char*, int*, int*, std::complex<double>*, int*, \
			  std::complex<double>* );
extern "C" void zgecon_( char*, int*, std::complex<double>*, int*, double*, \
			double*, std::complex<double>*, double*, int* );
extern "C" void zgetrf_( int*, int*, std::complex<double>*, int*, int*, int* );
extern "C" void zgetri_( int*, std::complex<double>*, int*, int*,	\
			std::complex<double>*, int* , int* );

extern "C" double dlange_(char*, int*, int*, double*, int*, double*);
extern "C" void dgecon_(char*, int*, double*, int*, double*,
			double*, double*, double*, int*);
extern "C" void dgetrf_(int*, int*, double*, int*, int*, int*);
extern "C" void dgetri_(int*, double*, int*, int*, double*, int* , int* );


wavefunc::wavefunc( )
{
	ConfigFile cfg( "Params.cfg" );

	nattemptmax = cfg.getValueOfKey<int>( "nattemptmax" );
	ConditionNumberMin = cfg.getValueOfKey<double>( "ConditionNumberMin" );
	BigNumber = 5000.0; 

	miscutils::InfoPrint( "Trial Wavefunction Initialization Parameters" );
	miscutils::InfoPrint( "BigNumber", BigNumber );
	miscutils::InfoPrint( "Minimum condition number", ConditionNumberMin );
	miscutils::InfoPrint( "Maximum generation attempts", nattemptmax );

	Delta = cfg.getValueOfKey<double>( "Delta" );
	mu = cfg.getValueOfKey<double>( "mu" );
	miscutils::InfoPrint( "Variational Parameters (Nondimensionalized by t)" );
	miscutils::InfoPrint( "Gap (Delta)", Delta );
	miscutils::InfoPrint( "Chemical potential (mu)", mu );
}

void wavefunc::SetUpPairFunc( sqlat &sq ) 
{
	double kx, ky;
	int nsites, i, j, k, r[2];
	double *A;
	double E, Delk, discE, cskx, csky, cs2kx, cs2ky, epsk, kdotr;

	nsites = sq.nsites;

	A = new double [nsites];

	for ( i = 0; i < nsites; i++ ) 
	{
		kx     = sq.kx[i];
		ky     = sq.ky[i];

		cskx = cos( kx );
		csky = cos( ky );
		cs2kx = cos( 2.0 * kx );
		cs2ky = cos( 2.0 * ky );

		epsk = 2.0 * ( sq.thop[0] * ( cskx + csky ) + 2.0 * sq.thop[1] * 
			cskx * csky + sq.thop[2] * ( cs2kx + cs2ky ) );
		E = epsk - mu;

		if ( ( fabs( kx ) < tol ) && ( fabs( ky ) < tol ) ) 
		{
			A[i] = 0.0;
			if ( E < 0.0 ) 
			{
				A[i] = BigNumber;
			}
		}
		else 
		{
			Delk  = sq.Jij * Delta * ( cskx - csky );
			discE = sqrt( E * E + Delk * Delk );
			A[i]  = -Delk / ( E + discE );
		}
	}

	// Now, set up the pairfunctions
	pairfn = new double [nsites*nsites];

	std::fill_n( pairfn, nsites * nsites, 0.0 ); 
	for ( i = 0; i < nsites; i++ ) 
	{
		for ( j = 0; j < nsites; j++ ) 
		{
                        sq.VecDistance( i, j, r );
			for ( k = 0; k < nsites; k++ ) 
			{
				kdotr = ( ( double ) r[0] ) * sq.kx[k] + ( ( double ) r[1] ) * sq.ky[k];
				pairfn[nsites*i+j] += ( cos( kdotr ) * A[k] );
			}
			pairfn[nsites*i+j] /= ( double ) nsites;
		}
	}
	delete A;
}

void wavefunc::SetUpInitialWavefunction( config& conf, int rand_seed )
{
	int nalloc, ispin, isite, i, j, nattempts;
	double *rwork;
	std::complex<double> *wk, *work, *psi; // Note: these were complex doubles in the serial C++ version
	double anorm, condition;
	int *ipiv;
	int info, lwork, nsites, npairs, nelecs;
	char norm[2] = "1";

	nsites = conf.nsites;
	npairs = conf.npairs;
	nelecs = 2 * npairs;

	miscutils::InitializeRandomNumGenerator( rand_seed );

	// Allocate temporary space
	lwork = npairs;
	psi = new std::complex<double> [npairs*npairs];
	wk = new std::complex<double> [nsites/2+(nsites/2-1)*nsites/4];
	work = new std::complex<double> [nelecs];
	ipiv = new int [npairs];
	rwork = new double [nelecs];

	nattempts = 0;

	while ( true )
	{
		nattempts++;
		if ( nattempts > nattemptmax )
		{
			miscutils::InfoPrint( "ERROR: Too many attempts in SetupInitialWaveFunction" );
			exit( 0 );
		}

		std::fill_n( conf.latocc, nsites, HL );

		// Generate Random Pairs
		for ( ispin = 0; ispin < 2; ispin++ )
		{
			nalloc = 0;
			while ( nalloc < npairs ) 
			{
				isite = miscutils::RandomInteger( nsites );
				if ( conf.latocc[isite] == HL ) 
				{
					conf.latocc[isite] = ispin;
					conf.whichpair[isite*2+ispin] = nalloc;
					conf.plist[nalloc*2+ispin] = isite;
					nalloc++;
				}
			}
		}

		// Now fill up the matrix psi
		for ( i = 0; i < npairs; i++ ) 
		{
			for ( j = 0; j < npairs; j++ ) 
			{
				// Store elements column-wise (Fortran style)
				psi[i+j*npairs] = pairfn[conf.plist[i*2+1]*nsites+conf.plist[j*2]];
				conf.PsiInv[i+j*npairs] = real( psi[i+j*npairs] );
			}
		}

		// Examine how well-conditioned the matrix is
		anorm = zlange_( norm, &npairs, &npairs, psi, &npairs, work );
		zgetrf_( &npairs, &npairs, psi, &npairs, ipiv, &info );
		if ( info != 0 )
		{
			miscutils::InfoPrint( "Error at zgetrf" ); 
		}
		zgecon_( norm, &npairs, psi, &npairs, &anorm, &condition, work, rwork, &info );
		if ( info != 0 ) 
		{
			miscutils::InfoPrint( "Error at zgecon" ); 
			exit( 0 );
		}
		if ( condition > ConditionNumberMin )
		{
			break;
		}
	};

	zgetri_( &npairs, psi, &npairs, ipiv, work, &lwork, &info );

	for ( i = 0; i < npairs; i++ )
	{
		for ( j = 0; j < npairs; j++ ) 
		{
			// Store elements row-wise in PsiInv now (i.e., C++ style)
			conf.PsiInv[i*npairs+j] = real( psi[i+j*npairs] );
			//cout << conf.PsiInv[i*npairs+j] << endl;
		}
	}

/*	miscutils::InfoPrint( "Initial Wavefunction Generation" );
	miscutils::InfoPrint( "Number of attempts", nattempts );
	miscutils::InfoPrint( "Condition number", condition ); */

	delete psi;
	delete wk;
	delete work;
	delete ipiv;
	delete rwork;
}

double wavefunc::VerifyInv
(config& conf, double *a_cmajor, double *b_rmajor, const char *msg)
{
  double err_sum = 0;
  const bool show_ident = false;
  const int npairs = conf.npairs;

  for(int i=0;i<npairs;i++) {
    for(int j=0;j<npairs;j++) {
      double aij = 0.0;
      for(int k=0;k<npairs;k++) {
        aij = aij + a_cmajor[k*npairs+i]*b_rmajor[k*npairs+j];
      }
      err_sum += fabs( i == j ? aij - 1 : aij );
      if ( show_ident )
        std::cout<<std::setw(6)<<std::setprecision(1)<<aij<<" ";
    }
    if ( show_ident )
      std::cout<<std::endl;
  }
  if ( false && msg )
    printf("Inv error for %s: %e\n",msg,err_sum);

  return err_sum;
}

double wavefunc::VerifyPsiInv(config& conf)
{
  const int nsites = conf.nsites;
  int npairs = conf.npairs;

  double* const psi = new double[npairs*npairs];
  double *a = new double[npairs*npairs];
  int *ipiv = new int[npairs];
  int lwork = npairs * 16;
  double *work  = new double[lwork];
  int info;

  for(int i=0;i<npairs;i++) {
    for(int j=0;j<npairs;j++) {
      //Store elements column-wise (Fortran style)
      int idx = i+j*npairs;
      psi[idx] = pairfn[conf.plist[i*2+1]*nsites+conf.plist[j*2]];
      a[idx] = psi[idx];
    }
  }

  const double err = VerifyInv(conf,psi,conf.PsiInv,"Existing PsiInv");

  if ( false )
    {
      dgetrf_(&npairs,&npairs,a,&npairs,ipiv,&info);
      dgetri_(&npairs,a,&npairs,ipiv,work,&lwork,&info);

      for(int i=0;i<npairs;i++)
        for(int j=0;j<npairs;j++)
          conf.PsiInv[i+j*npairs] = a[j+i*npairs]; 

      VerifyInv(conf,psi,conf.PsiInv,"Fresh PsiInv");
    }
  delete psi;
  delete a;
  delete ipiv;
  delete work;

  return err;
}
